from collections import OrderedDict
from typing import ClassVar

from sqlalchemy import String
from sqlmodel import Field

from gisaf.models.geo_models_base import GeoPolygonModel, GeoPointModel
from gisaf.models.project import Project
from gisaf.baskets import MiscGeomBasket
from gisaf.importers import GeoDataImporter
from gisaf.models.info_item import InfoItem

from ..metadata import avsm, avf, lb, tdc

#class CadastralOwnerType(Model):
#    name = db.Column(db.String, primary_key=True)
#t
#
#class CadastralUse(Model):
#    name = db.Column(db.String, primary_key=True)
#
#
#class CadastralCategory(Model):
#    name = db.Column(db.String, primary_key=True)
#
#
#class CadastralAvRelation(Model):
#    name = db.Column(db.String, primary_key=True)
#
#
#class CadastralArea(Model):
#    name = db.Column(db.String, primary_key=True)
#
#
#class CadastralPriority(Model):
#    name = db.Column(db.String, primary_key=True)


class Cadastral(GeoPolygonModel, table=True):
    __tablename__ = 'cadastral'
    __table_args__ = tdc.table_args

    class Admin:
        menu = 'Planning'

    error_on_import_bad_geometry_types: ClassVar[bool] = False
    viewable_role: ClassVar[str] = 'cadastral TDC'
    downloadable_role: ClassVar[str] = 'cadastral TDC download'
    z_index: ClassVar[str] = 415
    simplify: ClassVar[bool] = False

    skip_columns: ClassVar[set[str]] = set((
        'Area_Total',
        'CalcAreaAc',
        'Shape_Leng',
        'Shape_Area',
        'Perimeter',
    ))

    synonyms: ClassVar[dict[str, str]] = {
        'VILLAGE': 'Village',
        'STATUS': 'Status',
    }

    id: int | None = Field(default=None, primary_key=True)

    VilCode: str = Field(sa_type=String(5))
    Village: str = Field(sa_type=String(20))
    Division: int
    Sub_Div: str = Field(sa_type=String(5))
    Sub_Sub: str = Field(sa_type=String(15))
    Plot_Nr: str = Field(sa_type=String(20))
    Status: str = Field(sa_type=String(20))
    Area_Purch: float
    Puramboke_: str = Field(sa_type=String(20))
    LB_Pur_Acr: float
    Community: str = Field(sa_type=String(20))
    MP_Type: str = Field(sa_type=String(30))

    @property
    def caption(self):
        if self.Status == 'AVF':
            if self.Community:
                return '{self.Community} ({self.Plot_Nr})'.format(self=self)
            else:
                return '{self.Status} ({self.Plot_Nr})'.format(self=self)
        else:
            if self.Village:
                if self.Status:
                    return '{self.Status} - {self.Village} ({self.Plot_Nr})'.format(self=self)
                else:
                    return '{self.Village} ({self.Plot_Nr})'.format(self=self)
            else:
                return '{self.Status} ({self.Plot_Nr})'.format(self=self)

    async def get_feature_properties(self):
        ## DEPRECATED
        return {
            'class': self.Status
        }

    async def get_info(self) -> list[InfoItem]:
        info = await super(Cadastral, self).get_info()
        if self.Community:
            info.append(InfoItem(key='community', value=self.Community))
        info.append(InfoItem(key='status', value=self.Status))
        info.append(InfoItem(key='type', value=self.MP_Type))
        info.append(InfoItem(key='plot', value=self.Plot_Nr))
        info.append(InfoItem(key='source', value="L'Avenir d'Auroville"))
        return info


class CadastralAVF(GeoPolygonModel, table=True):
    __tablename__ = 'cadastral_avf'
    __table_args__ = avf.table_args

    class Admin:
        menu = 'Planning'

    skip_columns: ClassVar[set[str]] = set((
        'Area_Total',
        'CalcAreaAc',
        'CALC_AREA',
        'CALC_AREA_',
        'Calc_area',
        'Area',
        'AREA',
        'Shape_Leng',
        'Shape_Area',
        'Perimeter',
        'ORIG_FID',
        'OBJECTID',
    ))

    synonyms: ClassVar[dict[str, str]] = {
        'PLOT_NR': 'plot_nr',
        'Plot_Nr': 'plot_nr',
        'STATUS': 'status',
        'Status': 'status',
        'PURAMBOKE_': 'puramboke',
    }

    viewable_role: ClassVar[str] = 'cadastral AVF'
    downloadable_role: ClassVar[str] = 'cadastral AVF download'
    z_index: ClassVar[int] = 416
    simplify: ClassVar[bool] = False

    id: int | None = Field(default=None, primary_key=True)
    plot_nr: str = Field(sa_type=String(20))
    status: str = Field(sa_type=String(20))
    puramboke: str = Field(sa_type=String(20))
    project_id: int = Field(foreign_key='project.id')

    @property
    def caption(self):
        return '{status} ({self.plot_nr})'.format(self=self, status=self.status)

    @classmethod
    async def get_popup(cls, df):
        return 'Cadastral AVF: ' + df['status'] + ' (' + df['plot_nr'] + ')'

    @classmethod
    async def get_properties(cls, df):
        return {
            'class': df['status'].str.replace(' ', '-').str.replace('.', '-'),
            #'plot': df['plot_nr'],
            'label': df['plot_nr'],
        }

    def get_external_record_url(self):
        return f'https://acarat.auroville.org.in/land/surveynumber?sn={self.plot_nr}'

    async def get_feature_properties(self):
        ## DEPRECATED
        return {
            'class': self.status.replace(' ', '-').replace('.', '-'),
            'plot': self.plot_nr,
        }

    async def get_info(self) -> list[InfoItem]:
        info = await super(CadastralAVF, self).get_info()
        info.append(InfoItem(key='status', value=self.status))
        info.append(InfoItem(key='plot', value=self.plot_nr))
        if self.puramboke:
            info.append(InfoItem(key='puramboke', value=self.puramboke))
        info.append(InfoItem(key='source', value='Deoyani, for the Auroville Foundation'))
        return info

    attribution = 'AV Foundation'


class CadastralLB(GeoPolygonModel, table=True):
    __tablename__ = 'cadastral_lb'
    __table_args__ = lb.table_args

    class Admin:
        menu = 'Planning'

    skip_columns: ClassVar[set[str]] = set((
        'Area_Total',
        'CalcAreaAc',
        'CALC_AREA',
        'CALC_AREA_',
        'Calc_area',
        'Area',
        'AREA',
        'Shape_Leng',
        'Shape_Area',
        'Perimeter',
        'ORIG_FID',
        'OBJECTID',
    ))

    synonyms: ClassVar[dict[str, str]] = {
        'PLOT_NR': 'plot_nr',
        'Plot_Nr': 'plot_nr',
        'STATUS': 'status',
        'Status': 'status',
        'PURAMBOKE_': 'puramboke',
    }

    viewable_role: ClassVar[str] = 'cadastral LB'
    downloadable_role: ClassVar[str] = 'cadastral LB download'
    z_index: ClassVar[int] = 416
    simplify: ClassVar[bool] = False

    id: int | None = Field(default=None, primary_key=True)
    plot_nr: str = Field(sa_type=String(20))
    status: str = Field(sa_type=String(20))
    puramboke: str = Field(sa_type=String(20))
    project_id: int = Field(foreign_key='project.id')

    @property
    def caption(self):
        return '{status} ({self.plot_nr})'.format(self=self, status=self.status)

    @classmethod
    async def get_popup(cls, df):
        return 'Cadastral LB: ' + df['status'] + ' (' + df['plot_nr'] + ')'

    @classmethod
    async def get_properties(cls, df):
        return {
            'class': df['status'].str.replace(' ', '-').str.replace('.', '-'),
            #'plot': df['plot_nr'],
            'label': df['plot_nr'],
        }

    def get_external_record_url(self):
        return f'https://acarat.auroville.org.in/land/surveynumber?sn={self.plot_nr}'

    async def get_feature_properties(self):
        ## DEPRECATED
        return {
            'class': self.status.replace(' ', '-').replace('.', '-'),
            'plot': self.plot_nr,
        }

    async def get_info(self) -> list[InfoItem]:
        info = await super(CadastralLB, self).get_info()
        info.append(InfoItem(key='status', value=self.status))
        info.append(InfoItem(key='plot', value=self.plot_nr))
        if self.puramboke:
            info.append(InfoItem(key='puramboke', value=self.puramboke))
        info.append(InfoItem(key='source', value='Auroville Land Board'))
        return info

    attribution = 'AV Land Board'


class CadastralLBDissolved(GeoPolygonModel, table=True):
    __tablename__ = 'cadastral_lb_dissolved'
    __table_args__ = lb.table_args

    # CREATE TABLE other_avf.cadastral_avf_dissolved (geom geometry(Polygon, 4326), id serial NOT NULL, status character varying(20)) WITH (OIDS=FALSE);
    # CREATE INDEX idx_cadastral_avf_dissolved_geom ON other_avf.cadastral_avf_dissolved USING gist (geom);
    # ALTER TABLE other_avf.cadastral_avf_dissolved OWNER TO avgis;

    class Admin:
        menu = 'Planning'

    viewable_role: ClassVar[str] = 'cadastral LB dissolved'
    downloadable_role: ClassVar[str] = 'cadastral LB dissolved download'
    z_index: ClassVar[int] = 415
    simplify: ClassVar[bool] = False

    id: int | None = Field(default=None, primary_key=True)
    status: str = Field(sa_type=String(20))

    @classmethod
    async def get_popup(cls, df):
        return 'Cadastral Land Board (dissolved): ' + df['status']

    @classmethod
    async def get_properties(cls, df):
        return {
            'class': df['status'].str.replace(' ', '-').str.replace('.', '-'),
        }

    @property
    def caption(self):
        return '{status}'.format(self=self, status=self.status)

    async def get_feature_properties(self):
        ## DEPRECATED
        return {
            'class': self.status.replace(' ', '-').replace('.', '-')
        }

    async def get_info(self) -> list[InfoItem]:
        info = await super(CadastralAVFDissolved, self).get_info()
        info.append(InfoItem(key='status', value=self.status))
        info.append(InfoItem(key='source', value='Auroville Land Board'))
        return info

    attribution = 'AV Land Board'


class CadastralLocation(GeoPointModel, table=True):
    __tablename__ = 'cadastral_locations_avf'
    __table_args__ = avf.table_args

    class Admin:
        menu = 'Planning'

    viewable_role: ClassVar[str] = 'cadastral AVF dissolved'
    downloadable_role: ClassVar[str] = 'cadastral AVF dissolved'
    z_index: ClassVar[int] = 417
    attribution: ClassVar[str] = 'AV Foundation'

    synonyms: ClassVar[dict[str, str]] = {
        'NAME': 'name',
    }

    id: int | None = Field(default=None, primary_key=True)
    name: str = Field(sa_type=String(200))

    @classmethod
    async def get_popup(cls, df):
        return df['name']

    @property
    def caption(self):
        return '{self.name}'.format(self=self)

    async def get_info(self) -> list[InfoItem]:
        info = await super(CadastralLocation, self).get_info()
        info.append(InfoItem(key='name', value=self.name))
        info.append(InfoItem(key='source', value='Deoyani, for the Auroville Foundation'))
        return info


class CadastralAVFDissolved(GeoPolygonModel, table=True):
    __tablename__ = 'cadastral_avf_dissolved'
    __table_args__ = avf.table_args

    # CREATE TABLE other_avf.cadastral_avf_dissolved (geom geometry(Polygon, 4326), id serial NOT NULL, status character varying(20)) WITH (OIDS=FALSE);
    # CREATE INDEX idx_cadastral_avf_dissolved_geom ON other_avf.cadastral_avf_dissolved USING gist (geom);
    # ALTER TABLE other_avf.cadastral_avf_dissolved OWNER TO avgis;

    class Admin:
        menu = 'Planning'

    viewable_role: ClassVar[str] = 'cadastral AVF dissolved'
    downloadable_role: ClassVar[str] = 'cadastral AVF dissolved download'
    z_index: ClassVar[int] = 415
    simplify: ClassVar[bool] = False

    id: int | None = Field(default=None, primary_key=True)
    status: str = Field(sa_type=String(20))

    @classmethod
    async def get_popup(cls, df):
        return 'Cadastral AVF (dissolved): ' + df['status']

    @classmethod
    async def get_properties(cls, df):
        return {
            'class': df['status'].str.replace(' ', '-').str.replace('.', '-'),
        }

    @property
    def caption(self):
        return '{status}'.format(self=self, status=self.status)

    async def get_feature_properties(self):
        ## DEPRECATED
        return {
            'class': self.status.replace(' ', '-').replace('.', '-')
        }

    async def get_info(self) -> list[InfoItem]:
        info = await super(CadastralAVFDissolved, self).get_info()
        info.append(InfoItem(key='status', value=self.status))
        info.append(InfoItem(key='source', value='Deoyani, for the Auroville Foundation'))
        return info

    attribution = 'AV Foundation'


class CadastralAVFImporter(GeoDataImporter):
    def preprocess(self, gdf):
        gdf.status.fillna('Private', inplace=True)

    def get_model(self, file_record):
        return CadastralAVF


class CadastralAVFBasket(MiscGeomBasket):
    name = 'Cadastral AVF'
    model = CadastralAVF
    role = 'cadastral AVF'
    columns = ['name', 'time', 'project', 'import', 'delete']
    upload_fields = ['project']
    importer_class = CadastralAVFImporter
    projects = [
        'Cadastral AVF MP',
        'Cadastral AVF NOTMP',
    ]


class CadastralLBImporter(GeoDataImporter):
    def preprocess(self, gdf):
        gdf.status.fillna('Private', inplace=True)

    def get_model(self, file_record):
        return CadastralLB


class CadastralLBBasket(MiscGeomBasket):
    name = 'Cadastral LB'
    model = CadastralLB
    role = 'cadastral LB'
    columns = ['name', 'time', 'project', 'import', 'delete']
    upload_fields = ['project']
    importer_class = CadastralLBImporter
    projects = [
        'Cadastral LB',
    ]