from datetime import datetime, date
from io import BytesIO
from collections import defaultdict
from typing import ClassVar
import logging
import asyncio
import concurrent.futures

import pandas as pd
from sqlmodel import Field, Relationship
from openpyxl import load_workbook

from gisaf.models.geo_models_base import TimePlottableModel
from gisaf.baskets import Basket
from gisaf.importers import Importer, ImportError
from gisaf.utils import upsert_df, delete_df, postgres_upsert
from gisaf.plugins import DownloadCSVPlugin
from gisaf.models.info_item import InfoItem
from gisaf.models.admin import BasketImportResult

from avgs_gisaf_extra_models.models_base import Resident
from avgs_gisaf_extra_models.models_base import GeoPointModelWithResident, GeoPointWithLocationModel, Location
from avgs_gisaf_extra_models.metadata import misc, avsm

logger = logging.getLogger(__name__)

# raingauge_av_level_id_seq = db.Sequence('raingauge_av_level_id_seq')
# raingauge_av_id_seq = db.Sequence('raingauge_av_id_seq')
# raingauge_region_level_id_seq = db.Sequence('raingauge_region_level_id_seq')
# raingauge_region_id_seq = db.Sequence('raingauge_region_id_seq')

class RainGaugeAVLevel(TimePlottableModel, table=True):
    __tablename__ = 'raingauge_av_level'
    __table_args__ = misc.table_args

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'RainGaugeAVLevelsModelView'

    id: int | None = Field(default=None, primary_key=True)
    level: float
    trace: bool = False
    ref_id: int = Field(foreign_key='avsm_misc.raingauge_av.id')
    ref: 'RainGaugeAV' = Relationship(back_populates='levels')

    values: ClassVar[list[dict[str, str]]] = [
        {
            'name': 'level',
            'title': 'Level',
            'unit': 'mm/day',
            'agg': 'sum',
            'chartType': 'bar',
            'chartColor': 'blue',
        }
    ]

    def __repr__(self):
        return '<models.RainGaugeAVLevel {:%Y-%m-%d, %H:%M}: {:.1f}>'.format(self.time, self.level)


class RainGaugeAV(GeoPointModelWithResident, table=True):
    __tablename__ = 'raingauge_av'
    __table_args__ = misc.table_args

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'WithLevelsAndLocationModelView'

    id: int | None = Field(default=None, primary_key=True)
    since: str  ## Year, should change to date (default: 1/1/year, YYYY-MM-DD)
    levels: list[RainGaugeAVLevel] = Relationship(back_populates='ref')
    location: Location = Relationship()
    resident: Resident = Relationship()

    icon: ClassVar[str] = 'avsm_misc-raingauge'

    async def get_info(self) -> list[InfoItem]:
        return [InfoItem(key='source', value='CSR Geomatics Studio')]


class RainGaugeRegionLevel(TimePlottableModel, table=True):
    __tablename__ = 'raingauge_region_level'
    __table_args__ = misc.table_args

    id: int | None = Field(default=None, primary_key=True)
    level: float
    trace: bool = False
    ref_id: int = Field(foreign_key='avsm_misc.raingauge_region.id')
    ref: 'RainGaugeRegion' = Relationship(back_populates='levels')

    values: ClassVar[list[dict[str, str]]] = [
        {
            'name': 'level',
            'title': 'Level',
            'unit': 'mm/day',
            'agg': 'sum',
            'chartType': 'bar',
            'chartColor': 'blue',
        }
    ]

    def __repr__(self):
        return '<models.RainGaugeRegionLevel {:%Y-%m-%d, %H:%M}: {:.1f}>'.format(self.time, self.level)


class RainGaugeRegion(GeoPointWithLocationModel, table=True):
    __tablename__ = 'raingauge_region'
    __table_args__ = misc.table_args

    location: Location = Relationship()

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'WithLevelsModelView'

    id: int | None = Field(default=None, primary_key=True)
    since: date  ## Year, should change to date (default: 1/1/year, YYYY-MM-DD)
    in_charge: str
    levels: list[RainGaugeRegionLevel] = Relationship(back_populates='ref')
    #levels = db.relationship('RainGaugeRegionLevel', backref='raingauge_region')

    icon: ClassVar[str] = 'avsm_misc-raingauge'

    async def get_info(self) -> list[InfoItem]:
        return [InfoItem(key='source', value='CSR Geomatics Studio')]


class RainGaugeImporter(Importer):
    async def do_import(self, file, dry_run=False, keep_existing_nan=False) -> BasketImportResult:
        loop = asyncio.get_running_loop()

        ## Copied from import_to_db.rain_gauge_value(data, file_name)
        file_path = self.basket.base_dir / file.dir / file.name
        file_name = str(file_path)
        rows = []
        try:
            ## xlsx
            with concurrent.futures.ProcessPoolExecutor() as pool:
                workbook = await loop.run_in_executor(pool, load_workbook, file_name)
        except Exception as err:
            logger.exception(err)
            raise ImportError(
                f"Error importing speadcheet (is it a valid .xlsx file?: {err}"
            )
        else:
            ## Get the headers: raingauge_id and location_id
            header = workbook.worksheets[0]
            raingauge_id = int(header['B1'].value)
            # location_id = int(header['B2'].value)

            today = date.today()

            for worksheet in workbook.worksheets:
                ## Take only sheets with a year as title
                try:
                    year = int(worksheet.title)
                except ValueError:
                    continue

                for row in worksheet.iter_rows(min_row=4, max_row=34):
                    ## Filter empty rows (no id)
                    day = int(row[0].value)
                    for cell in row[1:]:
                        month = cell.col_idx - 1
                        if month > 12:
                            continue
                        if not cell.value:
                            continue
                        if cell.value == 'NA':
                            value = None
                            trace = False
                        elif cell.value == 'T':
                            ## Flag for a trace of rain
                            value = 0
                            trace = True
                        else:
                            try:
                                value = float(cell.value)
                                trace = False
                            except ValueError:
                                continue
                            except TypeError:
                                value = cell.value
                        try:
                            time = datetime(year=year, month=month, day=day, hour=8, minute=30)
                        except ValueError as err:
                            continue
                        if time.date() > today:
                            continue
                        rows.append([time, value, trace, raingauge_id])
        df = pd.DataFrame(rows, columns=['time', 'level', 'trace', 'ref_id']).dropna()
        ## TODO: Get the records existing in the DB,
        ## and keep in the df only those which are new or different
        df_db = await self.basket.model.get_df(where=self.basket.model.ref_id==raingauge_id)
        if keep_existing_nan:
            df_db.dropna(subset=['level'], inplace=True)
        #df_db.set_index('id', inplace=True)
        ## See: https://kanoki.org/2019/07/04/pandas-difference-between-two-dataframes/
        #no_change = pd.merge(df_db, df, how='inner', on=list(df.columns))

        ## TODO: move to get_df: convert automatically int indexes to Int64
        #if not isinstance(df_db.id.dtype, pd.Int64Dtype):
        #    df_db['id'] = df_db['id'].astype('Int64')

        details = defaultdict(int)

        change = df_db.reset_index().merge(df, how='outer', indicator=True)
        details['Unchanged'] = len(change[change._merge=='both'])
        change = change[change._merge!='both']

        ## Set time as index, which is the pivot for determining new, old, update records
        change.set_index('time', drop=True, inplace=True)
        ## Count the records with the same date
        change['count'] = change.groupby('time').size()
        ## Records with 2 same dates are updates
        to_update = change[change['count']==2]
        ## Records with only 1 date are new or old
        single = change[change['count']==1]
        to_add = single[single._merge=='right_only']
        to_delete = single[single._merge=='left_only']

        to_add.drop(columns=['id', '_merge', 'count'], inplace=True)
        to_add.reset_index(inplace=True)

        def _agg_update(group):
            group.loc[group._merge=='right_only', 'id'] = group.loc[group._merge=='left_only', 'id'].iloc[0]
            return group.loc[group._merge=='right_only']

        to_update_processed = to_update.groupby('time').apply(_agg_update).drop(columns=['_merge', 'count'])
        if not dry_run:
            await delete_df(to_delete, self.basket.model)
            await upsert_df(to_update_processed, self.basket.model)
            await upsert_df(to_add, self.basket.model)
        details['In file'] = len(df)
        details['In DB'] = len(df_db)
        details['Added'] = len(to_add)
        details['Deleted'] = len(to_delete)
        details['Updated'] = len(to_update_processed)
        return BasketImportResult(message='Import successful', details=details)


class RainGaugeAVLevelBasket(Basket):
    name = 'Raingauge AV'
    model = RainGaugeAVLevel
    importer_class = RainGaugeImporter


class RainGaugeRegionLevelBasket(Basket):
    name = 'Raingauge region'
    model = RainGaugeRegionLevel
    importer_class = RainGaugeImporter


download_all_as_csv = DownloadCSVPlugin(
    name='Raw raingauge readings as CSV',
    stores=[
        'avsm_misc.raingauge_av',
        'avsm_misc.raingauge_region'
    ],
    icon='storage',
)