from collections import OrderedDict
from math import isnan
from io import BytesIO
from pathlib import Path
from datetime import date, datetime
from typing import ClassVar, Any

from sqlalchemy import Text, Unicode
from sqlmodel import Field, Relationship
from babel.dates import format_date
from graphviz import Digraph # type: ignore[import-untyped]

# from gisaf.utils import get_joined_query
from gisaf.models.models_base import Model
from gisaf.config import conf
from gisaf.baskets import Basket
from gisaf.plugins import ActionPlugin
from gisaf.models.info_item import InfoItem
from gisaf.models.info import ActionResults

from ...models_base import Location, Resident, GeoPointZModelWithResident
from ...metadata import water, avsm

class WasteWaterTreatmentPlantType(Model, table=True):
    __tablename__ = 'waste_water_treatment_plant_type'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentPlantModelViewRelatedData'

    id: int | None = Field(default=None, primary_key=True)
    name: str

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<models.WasteWaterTreatmentPlantType {self.name:s}>'.format(self=self)


class WasteWaterTreatmentType(Model, table=True):
    __tablename__ = 'waste_water_treatment_type'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentTypeModelViewRelatedData'

    id: int | None = Field(default=None, primary_key=True)
    stage: str
    description: str
    color: str

    def __str__(self):
        return '{}: {}'.format(self.stage, self.description)


class WasteWaterTreatmentComponentType(Model, table=True):
    __tablename__ = 'waste_water_treatment_component_type'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentPlantModelViewRelatedData'

    id: int | None = Field(default=None, primary_key=True)
    name: str
    treatment_type_id: int = Field(foreign_key=water.table('waste_water_treatment_type.id'))
    treatment_type: WasteWaterTreatmentType = Relationship()
    health_checks: str = Field(sa_type=Text)

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<models.WasteWaterTreatmentComponentType {self.name:s}>'.format(self=self)

    @classmethod
    def selectinload(cls):
        return [
            cls.treatment_type
        ]

    # join_with: ClassVar[dict[str, Any]] = {
    #     'treatment_type': WasteWaterTreatmentType,
    # }


class WasteWaterTreatmentPlantDesigner(Model, table=True):
    __tablename__ = 'waste_water_treatment_plant_designer'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentPlantModelViewRelatedData'

    id: int | None = Field(default=None, primary_key=True)
    name: str
    organisation_company: str
    mobile: str
    phone: str
    email: str

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<models.WasteWaterTreatmentPlantDesigner {self.name:s}>'.format(self=self)


class WasteWaterTreatmentPlantCareTaker(Model, table=True):
    __tablename__ = 'waste_water_treatment_plant_care_taker'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentPlantModelViewRelatedData'

    id: int | None = Field(default=None, primary_key=True)
    name: str
    mobile: str
    phone: str
    email: str

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<models.WasteWaterTreatmentPlantCareTaker {self.name:s}>'.format(self=self)


class WasteWaterTreatmentPlantArchitect(Model, table=True):
    __tablename__ = 'waste_water_treatment_plant_architect'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentPlantModelViewRelatedData'

    id: int | None = Field(default=None, primary_key=True)
    name: str
    mobile: str
    phone: str
    email: str

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<models.WasteWaterTreatmentPlantArchitect {self.name:s}>'.format(self=self)


class WasteWaterTreatmentPlantContractor(Model, table=True):
    __tablename__ = 'waste_water_treatment_plant_contractor'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentPlantModelViewRelatedData'

    id: int | None = Field(default=None, primary_key=True)
    name: str
    contact_person: str
    mobile: str
    phone: str
    email: str

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<models.WasteWaterTreatmentPlantContractor {self.name:s}>'.format(self=self)


class WasteWaterTreatmentPlantStatus(Model, table=True):
    __tablename__ = 'waste_water_treatment_plant_status'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentPlantModelViewRelatedData'

    id: int | None = Field(default=None, primary_key=True)
    name: str

    def __str__(self):
        return self.name or ''

    def __repr__(self):
        return '<models.WasteWaterTreatmentPlantStatus {self.name:s}>'.format(self=self)


class WasteWaterTreatmentPlant(GeoPointZModelWithResident, table=True):
    __tablename__ = 'waste_water_treatment_plant'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentPlantModelView'

    id: int | None = Field(default=None, primary_key=True)
    user_t_id: int = Field(foreign_key=water.table('waste_water_treatment_plant_type.id'))
    status_id: int = Field(foreign_key=water.table('waste_water_treatment_plant_status.id'))
    notes: str
    designr_id: int = Field(foreign_key=water.table('waste_water_treatment_plant_designer.id'))
    architc_id: int = Field(foreign_key=water.table('waste_water_treatment_plant_architect.id'))
    caretkr_id: int = Field(foreign_key=water.table('waste_water_treatment_plant_care_taker.id'))
    contrac_id: int = Field(foreign_key=water.table('waste_water_treatment_plant_contractor.id'))
    inst_date: date
    dsludge_dt: date
    remarks: str = Field(sa_type=Text)
    problems: str = Field(sa_type=Text)
    modifications_history: str = Field(sa_type=Text)
    plan_cap: float
    plan_cap_p: int
    docs_collected: str = Field(sa_type=Text)  ## Attachments?
    ## Those will probably evolve:
    picture: str = Field(sa_type=Unicode(128))  ## Attachments
    latest_health_check: str = Field(sa_type=Unicode(128))  ## Attachments

    resident: Resident = Relationship()
    location: Location = Relationship()
    user_type: WasteWaterTreatmentPlantType = Relationship()
    wwtp_status: WasteWaterTreatmentPlantStatus = Relationship()
    designer: WasteWaterTreatmentPlantDesigner = Relationship()
    architect: WasteWaterTreatmentPlantArchitect = Relationship()
    care_taker: WasteWaterTreatmentPlantCareTaker = Relationship()
    contractor: WasteWaterTreatmentPlantContractor = Relationship()

    icon: ClassVar[str] = 'avsm_water-waste_water_treatment_plant'

    @classmethod
    def selectinload(cls):
        return [
            cls.resident,
            cls.location,
            cls.wwtp_status,
            cls.user_type,
            cls.designer,
            cls.architect,
            cls.contractor
        ]

    # @classmethod
    # def dyn_join_with(cls):
    #     return {
    #         'location': Location,
    #         'user_type': WasteWaterTreatmentPlantType,
    #         'status': WasteWaterTreatmentPlantStatus,
    #         'designer': WasteWaterTreatmentPlantDesigner,
    #         'resident': Resident.on(cls.res_id == Resident.id),
    #         'architect': WasteWaterTreatmentPlantArchitect,
    #         'care_taker': WasteWaterTreatmentPlantCareTaker,
    #         'contractor': WasteWaterTreatmentPlantContractor,
    #     }

    @property
    def caption(self):
        if self.status_id:
            return '{self.location} ({self.status}, {self.user_type}, id={self.id:d})'.format(self=self)
        elif self.user_t_id:
            return '{self.location} ({self.user_type}, id={self.id:d})'.format(self=self)
        else:
            return '{self.location} (id={self.id:d})'.format(self=self)

    @classmethod
    async def get_popup(cls, df):
        return 'Waste water treatment plant: ' + df.avsm_location_name + ' (' + \
            df.avsm_water_waste_water_treatment_plant_status_name + ', ' + \
            df.avsm_water_waste_water_treatment_plant_type_name + ', ' + \
            df.index.astype('U') + ')'

    @classmethod
    async def get_properties(cls, df):
        return {
            'class': df.avsm_water_waste_water_treatment_plant_status_name.combine_first(
                df.avsm_water_waste_water_treatment_plant_status_name
            ).str.replace(' ', '-')
        }

    async def get_info(self):
        info = OrderedDict(await super(WasteWaterTreatmentPlant, self).get_info())
        info['user type'] = self.user_type.name if self.user_t_id and self.user_type else None
        info['status'] = self.status.name if self.status_id else None
        info['last desluge'] = format_date(self.dsludge_dt, format='short') if self.dsludge_dt else None
        info['designer'] = self.designer.name if self.designr_id and self.designer else None
        info['installation date'] = format_date(self.inst_date, format='short') if self.inst_date else None
        if self.res_id:
            info['in charge'] = self.resident.name if self.resident else None
        info['capacity (m3/day)'] = self.plan_cap if self.plan_cap and not isnan(self.plan_cap) else None
        info['capacity (people)'] = self.plan_cap_p

        origin_types = await self.get_origin_types()
        if origin_types:
            info['waste water origin types'] = origin_types
        info['source'] = 'CSR Geomatics Studio'
        return info

    async def get_origin_types(self):
        """
        Return the set of origin types
        """
        components = await self.get_components()
        origins = await self.get_origin_links(components)
        origin_type_ids = set(oo.from_origin.type_id for oo in origins)
        origin_types = await WasteWaterTreatmentOriginType.query.where(WasteWaterTreatmentOriginType.id.in_(origin_type_ids)).gino.all()
        if origin_types:
            return ', '.join([type.name or "Unknown" for type in origin_types])

    async def get_origin_links(self, components):
        origin_links = set()
        model = WasteWaterTreatmentOriginToComponentFlow
        for component in components:
            query = get_joined_query(model).where(model.to_component_id == component.id)
            for link in await query.gino.all():
                origin_links.add(link)
        return sorted(origin_links, key=lambda x:x.id)

    async def get_output_links(self, components):
        output_links = set()
        model = WasteWaterTreatmentComponentToOutputFlow
        for component in components:
            query = get_joined_query(model).where(model.from_component_id == component.id).order_by(model.id).order_by(model.from_component_id)
            for link in await query.gino.all():
                output_links.add(link)
        return sorted(output_links, key=lambda x:x.id)

    async def get_components(self):
        model = WasteWaterTreatmentComponent
        components = await get_joined_query(model).where(model.plant_id == self.id).order_by(model.type_id, model.id).gino.all()
        return components

    async def get_links(self, components):
        links = {}
        for component in components:
            model = WasteWaterTreatmentComponentFlow
            query = get_joined_query(model).where((model.from_component_id == component.id)
                                                  | (model.to_component_id == component.id))
            for link in await query.gino.all():
                links[link.id] = link
        return set(links.values())


    class Attachments:
        async def files(self):
            if self.latest_health_check:
                return {
                    'latest_health_check': self.latest_health_check
                }

        async def images(self):
            if self.picture:
                return {
                    'picture': self.picture
                }

    async def get_graph(self, format='svg'):
        components = await self.get_components()
        if not components:
            return

        origin_links = await self.get_origin_links(components)
        output_links = await self.get_output_links(components)

        graph_attr = {
            'rankdir': 'LR',
            'size': "15,40",
            'nodesep': "0.15",
        }
        node_attr = {
            'fontname': 'Helvetica',
            'fontsize': '10',
            'shape': 'record',
            'style': 'rounded,filled',
            'height': '.2',
            'color': 'LightSteelBlue',
        }

        edge_attr = {
            'color': 'MediumSlateBlue'
        }

        dot = Digraph(
            engine='dot',
            graph_attr=graph_attr,
            node_attr=node_attr,
            edge_attr=edge_attr,
        )

        ## Nodes
        for comp in components:
            dot.node(str(comp.id), comp.label,
                     tooltip=comp.tooltip, fillcolor=comp.color)

        for origin_link in origin_links:
            dot.node(origin_link.origin_graph_id, origin_link.from_origin.label,
                     tooltip=origin_link.from_origin.tooltip,
                     fillcolor='white', shape='cds', height='0.4', color='black')

        for output_link in output_links:
            dot.node(output_link.output_graph_id, output_link.to_output.label,
                     tooltip=output_link.to_output.tooltip,
                     fillcolor='white', shape='box3d', height='0.4', color='black')

        ## Links
        for link in await self.get_links(components):
            dot.edge(str(link.from_component_id),
                     str(link.to_component_id),
                     tooltip='Pipe #{:d}'.format(link.id))
        for origin_link in origin_links:
            dot.edge(origin_link.origin_graph_id,
                     str(origin_link.to_component_id),
                     tooltip='Input pipe #{:d}'.format(origin_link.id))
        for output_link in output_links:
            dot.edge(str(output_link.from_component_id),
                     output_link.output_graph_id,
                     tooltip='Output pipe #{:d}'.format(output_link.id))

        graph = dot.pipe(format=format)
        if format == 'svg':
            return ''.join(graph.decode('utf-8').split('\n')[6:-1])
        else:
            return graph

    async def get_feature_properties(self):
        ## DEPRECATED
        if self.status_id:
            return {
                'class': self.status.name.replace(' ', '-')
            }
        elif self.user_t_id:
            return {
                'class': self.user_type.name.replace(' ', '-')
            }
        else:
            return {}


class WasteWaterTreatmentComponent(Model, table=True):
    __tablename__ = 'waste_water_treatment_component'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentComponentModelView'

    id: int | None = Field(default=None, primary_key=True)
    type_id: int = Field(foreign_key=water.table('waste_water_treatment_component_type.id'))
    type: WasteWaterTreatmentComponentType = Relationship()
    plant_id: int = Field(foreign_key=water.table('waste_water_treatment_plant.id'))
    plant: WasteWaterTreatmentPlant = Relationship()
    accessible: bool
    visible: bool
    remarks: str = Field(sa_type=Text)

    def __str__(self):
        if self.plant:
            return '{self.type} (id {self.id}) of {self.plant.caption}'.format(self=self)
        else:
            return '{self.type} (id {self.id})'.format(self=self)

    def __repr__(self):
        return '<models.WasteWaterTreatmentComponent {self.type} {self.plant.caption:s}>'.format(self=self)

    @classmethod
    def selectinload(cls):
        return [
            cls.plant,
            cls.type
        ]

    # join_with: ClassVar[dict[str, Any]] = {
    #     'plant': WasteWaterTreatmentPlant,
    #     'type': WasteWaterTreatmentComponentType,
    # }

    @property
    def label(self):
        return '{self.type} (#{self.id})'.format(self=self)

    @property
    def tooltip(self):
        if self.type.treatment_type.description:
            return '{self.type.treatment_type.stage}: {self.type.treatment_type.description}'.format(self=self)
        else:
            return '{self.type.treatment_type.stage}'.format(self=self)

    @property
    def color(self):
        return self.type.treatment_type.color or ''


class WasteWaterTreatmentComponentFlow(Model, table=True):
    __tablename__ = 'waste_water_treatment_component_flow'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentComponentFlowModelView'

    id: int | None = Field(default=None, primary_key=True)
    from_component_id: int = Field(foreign_key=water.table('waste_water_treatment_component.id'))
    to_component_id: int = Field(foreign_key=water.table('waste_water_treatment_component.id'))

    def __str__(self):
        return '{self.from_component} - {self.to_component}'.format(self=self)

    def __repr__(self):
        return '<models.WasteWaterTreatmentComponentFlow {self.id:d}>'.format(self=self)

    @classmethod
    def dyn_join_with(cls):
        WasteWaterTreatmentComponentTo = WasteWaterTreatmentComponent.alias()
        return {
            'from_component': WasteWaterTreatmentComponent.on(
                cls.from_component_id == WasteWaterTreatmentComponent.id),
            'to_component': WasteWaterTreatmentComponentTo.on(
                cls.to_component_id == WasteWaterTreatmentComponentTo.id),
        }


class WasteWaterTreatmentOriginType(Model, table=True):
    __tablename__ = 'waste_water_treatment_origin_type'
    __table_args__ = water.table_args

    ## CREATE TABLE avsm_water.waste_water_treatment_origin_type (id serial NOT NULL, name character varying, CONSTRAINT waste_water_treatment_origin_type_pkey PRIMARY KEY (id)) WITH (OIDS=FALSE);

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentPlantModelViewRelatedData'

    id: int | None = Field(default=None, primary_key=True)
    name: str

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<models.WasteWaterTreatmentOriginType {self.name:s}>'.format(self=self)


class WasteWaterTreatmentOrigin(Model, table=True):
    __tablename__ = 'waste_water_treatment_origin'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentPlantModelViewRelatedData'

    id: int | None = Field(default=None, primary_key=True)
    name: str
    #plant_id = db.Column(db.Integer, db.ForeignKey(WasteWaterTreatmentPlant.id))
    plant_id: int = Field(foreign_key=water.table('waste_water_treatment_plant.id'))
    plant: WasteWaterTreatmentPlant = Relationship()
    loc_id: int = Field(foreign_key=avsm.table('location.id'))
    location: Location = Relationship()
    type_id: int = Field(foreign_key=water.table('waste_water_treatment_origin_type.id'))
    type: WasteWaterTreatmentOriginType = Relationship()

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<models.WasteWaterTreatmentOrigin {self.name:s}>'.format(self=self)

    @classmethod
    def selectinload(cls):
        return [
            cls.plant,
            cls.location,
            cls.type,
        ]

    # join_with: ClassVar[dict[str, Any]] = {
    #     'plant': WasteWaterTreatmentPlant,
    #     'location': Location,
    #     'type': WasteWaterTreatmentOriginType,
    # }

    @property
    def label(self):
        return self.name

    @property
    def tooltip(self):
        if self.type_id:
            if self.name != self.location.name:
                return 'Origin: {self.name} ({self.type}) in {self.location.name}'.format(self=self)
            else:
                return 'Origin: {self.name} ({self.type})'.format(self=self)
        else:
            if self.name != self.location.name:
                return 'Origin: {self.name} in {self.location.name}'.format(self=self)
            else:
                return 'Origin: {self.name}'.format(self=self)


class WasteWaterTreatmentOutputCategory(Model, table=True):
    __tablename__ = 'waste_water_treatment_output_category'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentPlantModelViewRelatedData'

    id: int | None = Field(default=None, primary_key=True)
    name: str

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<models.WasteWaterTreatmentOutputCategory {self.name:s}>'.format(self=self)


class WasteWaterTreatmentOutput(Model, table=True):
    __tablename__ = 'waste_water_treatment_output'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentPlantModelViewRelatedData'

    id: int | None = Field(default=None, primary_key=True)
    name: str
    #plant_id = db.Column(db.Integer, db.ForeignKey(WasteWaterTreatmentPlant.id))
    loc_id: int = Field(foreign_key=avsm.table('location.id'))
    location: Location = Relationship()
    cat_id: int = Field(foreign_key=water.table('waste_water_treatment_output_category.id'))
    category: WasteWaterTreatmentOutputCategory = Relationship()

    def __str__(self):
        return '{self.name} ({self.category} in {self.location})'.format(self=self)

    def __repr__(self):
        return '<models.WasteWaterTreatmentOutput {self.name:s}>'.format(self=self)

    @classmethod
    def selectinload(cls):
        return [
            cls.location,
            cls.category,
        ]

    # join_with: ClassVar[dict[str, Any]] = {
    #     'location': Location,
    #     'category': WasteWaterTreatmentOutputCategory,
    # }

    @property
    def label(self):
        return self.category.name

    @property
    def tooltip(self):
        if self.name:
            return 'Output: {self.name} in {self.location.name}'.format(self=self)
        else:
            return 'Output: {self.category.name} in {self.location.name}'.format(self=self)


class WasteWaterTreatmentOriginToComponentFlow(Model, table=True):
    __tablename__ = 'waste_water_treatment_origin_to_component_flow'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentOriginToComponentFlowModelView'

    id: int | None = Field(default=None, primary_key=True)
    from_origin_id: int = Field(foreign_key=water.table('waste_water_treatment_origin.id'))
    from_origin: WasteWaterTreatmentOrigin = Relationship()
    to_component_id: int = Field(foreign_key=water.table('waste_water_treatment_component.id'))
    to_component: WasteWaterTreatmentComponent = Relationship()

    def __str__(self):
        return '{self.from_origin} - {self.to_component}'.format(self=self)

    def __repr__(self):
        return '<models.WasteWaterTreatmentOriginToComponentFlow {self.id:d}>'.format(self=self)

    @classmethod
    def selectinload(cls):
        return [
            cls.from_origin,
            cls.to_component,
        ]

    # join_with: ClassVar[dict[str, Any]] = {
    #     'from_origin': WasteWaterTreatmentOrigin,
    #     'to_component': WasteWaterTreatmentComponent,
    # }

    @property
    def origin_graph_id(self):
        return 'o_{:d}'.format(self.from_origin_id)


class WasteWaterTreatmentComponentToOutputFlow(Model, table=True):
    __tablename__ = 'waste_water_treatment_component_to_output_flow'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water treatment'
        flask_admin_model_view = 'WasteWaterTreatmentComponentToOutputFlowModelView'

    id: int | None = Field(default=None, primary_key=True)
    from_component_id: int = Field(foreign_key=water.table('waste_water_treatment_component.id'))
    from_component: WasteWaterTreatmentComponent = Relationship()
    to_output_id: int = Field(foreign_key=water.table('waste_water_treatment_output.id'))
    to_output: WasteWaterTreatmentOutput = Relationship()

    def __str__(self):
        return '{self.from_component} - {self.to_output}'.format(self=self)

    def __repr__(self):
        return '<models.WasteWaterTreatmentComponentToOutputFlow{self.id:d}>'.format(self=self)

    @classmethod
    def selectinload(cls):
        return [
            cls.from_component,
            cls.to_output,
        ]

    # join_with: ClassVar[dict[str, Any]] = {
    #     'from_component': WasteWaterTreatmentComponent,
    #     'to_output': WasteWaterTreatmentOutput,
    # }

    @property
    def output_graph_id(self):
        return 'v_{:d}'.format(self.to_output_id)


class TemplateBasket(Basket):
    name = 'WWTP report template'


class GetReportTemplate(ActionPlugin):
    async def execute(self, user, features, params) -> ActionResults:
        raise NotImplementedError('This function is currently disabled. Please contact us.')
        ## TODO: download file - custom download plugin?

        from docxtpl import DocxTemplate, InlineImage
        from docx.shared import Mm

        registry = request.app['registry']

        now = datetime.now()

        ## Get template
        ## FIXME: test (was conf.template_import_path => cond.admin['basket']...
        ## TODO: response with a proper error if the template is not found
        template_path = Path(conf.admin['basket']['base_dir'])/'WWTP report template'/'wwtp_health_check_template.docx'
        with open(template_path, 'rb') as f:
            source_stream = BytesIO(f.read())

        doc = DocxTemplate(source_stream)

        for store, ids in features.items():
            model = registry.stores.loc[store, 'model']
            #items = await model.get_df(where=model.id.in_(ids))
            for id in ids:
                wwtp = await get_joined_query(model).where(model.id == id).gino.first()

                components = await wwtp.get_components()
                origins = await wwtp.get_origin_links(components)
                outputs = await wwtp.get_output_links(components)
                origin_types = await wwtp.get_origin_types()
                ## Hack for health checks, separated by '\n'
                for component in components:
                    if component.type.health_checks:
                        component.health_checks = component.type.health_checks.strip().split('\n')
                    else:
                        component.health_checks = []
                context = {
                    'system' : wwtp,
                    'now': now,
                    'components': components,
                    'origins': origins,
                    'origin_types': origin_types,
                    'outputs': outputs,
                }

                if components:
                    graph_png = await wwtp.get_graph(format='png')
                    context['picture'] = InlineImage(doc, BytesIO(graph_png), width=Mm(180))

                doc.render(context)

                target_stream = BytesIO()
                doc.save(target_stream)
                target_stream.seek(0)

                file_name = 'Health check report - {} (id {} - template {:%Y-%m-%d}).docx'.format(
                    wwtp.location.name,
                    wwtp.id,
                    now,
                )

                return {
                    'file_name': file_name,
                    'content': target_stream,
                    'content_type': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
                }


get_report_template = GetReportTemplate(
    name='Get an empty report for health check',
    stores=['avsm_water.waste_water_treatment_plant'],
)

