import logging
from collections import OrderedDict
from datetime import datetime, date as date_
# import dateutil.parser  # TODO: find a workaround for dateutil python deps issue with pdm
import dateutil
from tzlocal import get_localzone
from typing import ClassVar, Literal, Any

from sqlalchemy import String
from sqlalchemy.dialects.postgresql import HSTORE
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import QueryableAttribute, selectinload, joinedload
from sqlalchemy.sql.expression import func
from sqlmodel import Field, Relationship, select, SQLModel
import pandas as pd

from gisaf.database import db_session
from gisaf.baskets import Basket
from gisaf.plugins import (ActionPlugin, DownloadCSVPlugin,
                           UserRead, FormFieldInput)
from gisaf.models.models_base import Model
from gisaf.models.geo_models_base import TimePlottableModel
from gisaf.models.misc import Qml
from gisaf.models.info import (InfoCategory, InfoItem, Attachment, FormField,
                               PlotParams, PlotBaseLine, PlotBgShape,
                               ActionResult)

from avgs_gisaf_extra_models.models_base import (
    GeoPointZModelWithResident, Location, Resident, InfoSource)

from ...geology.formation import GeologicalFormationType

from ...metadata import water

logger = logging.getLogger(__name__)

# well_type_id_seq = db.Sequence('well_type_id_seq')
# well_status_id_seq = db.Sequence('well_status_id_seq')
# well_id_seq = db.Sequence('well_id_seq')
# well_level_id_seq = db.Sequence('well_level_id_seq')

class NoCasingTop(Exception):
    pass

class NoLevel(Exception):
    pass


removable_prefix = 'avsm_water_well_'
def _pretty_print_key(k):
    """
    Utility function to transform a string k to human friendly format,
    used for keys of additional well info
    """
    if k.startswith(removable_prefix):
        unprefixed = k[len(removable_prefix):]
    else:
        unprefixed = k
    return unprefixed.replace('_', ' ')


class WellMonitoringStatus(Model, table=True):
    __tablename__ = 'well_monitoring_status'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'MyModelViewWithName'

    id: int | None = Field(default=None, primary_key=True)
    name: str = Field(sa_column_kwargs={'nullable': False, 'unique': True})

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<models.WellMonitoringStatus {self.name:s}>'.format(self=self)


class WellType(Model, table=True):
    __tablename__ = 'well_type'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'MyModelViewWithName'

    id: int | None = Field(default=None, primary_key=True)
    name: str

    def __str__(self):
        return self.name or 'well type missing'

    def __repr__(self):
        return '<models.WellType {self.name:s}>'.format(self=self)


class PumpStatus(Model, table=True):
    __tablename__ = 'pump_status'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'MyModelViewWithName'

    id: int | None = Field(default=None, primary_key=True)
    name: str


class WellStatus(Model, table=True):
    __tablename__ = 'well_status'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'MyModelViewWithName'

    id: int | None = Field(default=None, primary_key=True)
    name: str

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<models.WellStatus {self.name:s}>'.format(self=self)


class Well(GeoPointZModelWithResident, table=True):
    __tablename__ = 'well'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'WellModelView'

    id: int | None = Field(default=None, primary_key=True)
    wellcode: str
    type_id: str = Field(foreign_key='avsm_water.well_type.id')
    status_id: int = Field(foreign_key='avsm_water.well_status.id')
    moni_st_id: int = Field(foreign_key='avsm_water.well_monitoring_status.id')
    infosrc_id: int = Field(foreign_key='avsm.info_source.id')
    cf: float  ## Correction factor
    cf_hdpe_pipe: float  ## Length of the HDPE pipe for measuring
    depth: float
    casing_top: float  ## Reference to the sea level ( = elevation + cf)
    notes: str
    obs: str  ## Observation
    picture: str = Field(sa_type=String(128))
    location: Location = Relationship()
    resident: Resident = Relationship()
    well_status: WellStatus = Relationship()
    well_type: WellType = Relationship()
    monitoring_status: WellMonitoringStatus = Relationship()
    info_source: InfoSource = Relationship()
    # pump_status: PumpStatus = Relationship()
    # levels: list['WellLevel'] = Relationship()
    # location: Location = Relationship()
    # location: Location = Relationship(sa_relationship_kwargs={
    #     #  'primaryjoin': 'Well.loc_id == Location.id',
    #      'lazy': 'dynamic'
    # })

    ## XXX: Notes for ticket #10569
    # - OBJECT-ID: probably an artifact, skip


#    ## More fields, refs #10569
#    accuracy_b = db.Column(db.Float)
#    remarks = db.Column(db.String)
#    h_wellcode = db.Column(db.String(16))
#    h_codepres = db.Column(db.Boolean)
#    descr_h = db.Column(db.String)
#    codeonwellb = db.Column(db.String(20))
#    access_b = db.Column(db.Boolean)
#    drill_depth = db.Column(db.Float)
#    depth_b = db.Column(db.Float)
#    ## XXX: casing fields: which field missing in existing?
#    pump_b = db.Column(db.Boolean)
#    pump_status_b_id = db.Column(db.Integer, db.ForeignKey(PumpStatus.id)))
#    pump_depth_b =

    icon: ClassVar[str] = 'avsm_water-well'
    symbol: ClassVar[str] = '\ue32e'
    cache_enabled: ClassVar[bool] = False

    @classmethod
    def selectinload(cls) -> list[Literal['*'] | QueryableAttribute[Any]]:
        return [
            cls.location,
            cls.resident,  # Resident.on(cls.res_id == Resident.id),
            cls.well_type,  # WellType,
            cls.well_status,
            cls.monitoring_status,
            cls.info_source,
            # cls.pump_status,
        ]

    @classmethod
    async def get_legend(cls):
        return OrderedDict([
            ('Blue', 'Monitored, functioning'),
            ('Red', 'Dry'),
            ('Yellow', 'Level under see level'),
            ('Brown', 'Not functioning'),
            ('Grey', 'Not monitored'),
        ])

    @classmethod
    async def get_popup(cls, df):
        return 'Well: ' + df.avsm_location_name \
            + ' #' + df.index.astype('str') \
            + ' (' + df.wellcode + ', ' \
                + df.avsm_water_well_status_name.fillna('no status') \
                + ', ' + df.avsm_water_well_type_name.fillna('no type') + ')'

    @classmethod
    async def get_properties(cls, df):
        ## XXX: Should get the latest levels per ref_id in the DB request
        ##      For the time being, get all the levels
        ##      and grouping/lastest level per well in the dataframe
        levels = await WellLevel.get_df()
        levels.dropna(subset=['ref_id'], inplace=True)
        levels.set_index(levels.time, inplace=True)
        levels.sort_index(inplace=True)
        last_levels = levels.groupby(levels.ref_id).nth(-1)
        df_with_latest_levels = df.merge(
            last_levels[['ref_id', 'level', 'dry']],
            left_index=True,
            right_on='ref_id',
            how='left').set_index('ref_id')
        def get_class(well_level):
            if well_level.avsm_water_well_monitoring_status_id == 2 \
                or pd.isna(well_level.depth):
                return 'notMonitored'
            if well_level.status_id != 1:
                return 'notFunctioning'
            if well_level.dry or (well_level.level <= -well_level.depth):
                return 'dry'
            if well_level.level <= -well_level.geom.z:
                return 'underSeaLevel'
            return 'functioning'
        return {
            ## XXX: using apply, could be vectorized
            'class': df_with_latest_levels.apply(get_class, axis=1)
        }

    @property
    def caption(self):
        return '{self.location} ({self.wellcode}, {type}, {status}, id={self.id:d})'.\
            format(self=self,
                   status=self.well_status.name if self.status_id else 'status unknown',
                   type=self.well_type.name if self.type_id else 'type unknown',
                  )

    async def get_feature_properties(self):
        ## DEPRECATED
        if self.moni_st_id.id == 2:  ## Not monitored
            return {
                'class': 'notMonitored'
            }
        if self.status_id != 1:  ## Functioning/in use
            return {
                'class': 'notFunctioning'
            }
        measure = await WellLevel.query.where(WellLevel.ref_id==self.id).order_by(WellLevel.time.desc()).gino.first()
        if measure and measure.level:
            if measure.dry or (self.depth and measure.level <= -self.depth):
                return {
                    'class': 'dry'
                }
            if self.elevation and measure.level <= -self.elevation:
                return {
                    'class': 'underSeaLevel'
                }
            else:
                return {
                    'class': 'functioning'
                }
        else:
            return {
                'class': 'functioning'
            }

    def get_abs_level(self):
        """
        Get the absolute level and time (last reading)
        :return:
        """
        measure = self.levels.order_by(WellLevel.time.desc()).first()
        return (measure.level + self.casing_top, measure.time)

    async def get_trend(self):
        async with db_session() as session:
            query = select(WellLevel).where(WellLevel.ref_id==self.id).\
                order_by(WellLevel.time.desc()).limit(2)
            result = await session.exec(query)
            last_2_levels = result.all()
        # last_2_levels = await WellLevel.query.where(WellLevel.ref_id==self.id).\
        #     order_by(WellLevel.time.desc()).limit(2).gino.all()
        if len(last_2_levels) == 2:
            last_level_item = last_2_levels[0]
            last_level = last_level_item.level
            try:
                level_info = "{:.3f}m".format(last_level_item.as_absolute(self))
            except NoCasingTop:
                level_info = "{:.3f}m (no casing top)".format(last_level_item.level)
            level_time = "{:%Y-%m-%d, %Hh %Mm}".format(last_level_item.time)

            try:
                trend = last_2_levels[0].level - last_2_levels[1].level
                interval = last_2_levels[0].time - last_2_levels[1].time
            except TypeError:
                trend = "Level missing"
            else:
                ## TODO: render as red if the trend is negative, as in this old piece of code:
                #resp['info']['trend'] = "<p{:s}>Trend: {:+.1f}m in {:d}day(s), {:d}h {:d}m</p>".format(
                #    " class='red'" if trend < 0 else '',
                #    float(trend),
                #    interval.days, interval.seconds // 3600, interval.seconds // 60)
                trend = "{:+.2f}m in {:d}day(s)".format(float(trend), interval.days)
        elif len(last_2_levels) == 1:
            last_level = last_2_levels[0].level
            level_info = last_level
            trend = 'Only one reading'
            level_time = "{:%Y-%m-%d, %Hh %Mm}".format(last_2_levels[0].time)
        else:
            last_level = None
            level_info = None
            trend = None
            level_time = None
        return level_info, trend, level_time, last_level

    @classmethod
    async def get_plot_params(cls, well_id, value) -> PlotParams:
        well_id = int(well_id)
        ## Depth and bottom lines
        well_df = await cls.get_geo_df(where=cls.id==well_id)
        well = well_df.iloc[0]
        plot_params = PlotParams()
        is_amsl = value == 'level_amsl'
        if is_amsl:
            amsl_shift = float(well.geom.z) \
                + (0 if pd.isna(well.casing_top) else well.casing_top) \
                + (0 if pd.isna(well.cf) else well.cf)
        else:
            amsl_shift = 0
        if not pd.isna(well.depth):
            plot_params.baseLines.append(PlotBaseLine(
                name='Bottom',
                value=amsl_shift - float(well.depth),
                color='#202000'
            ))
        plot_params.baseLines.append(PlotBaseLine(
            name='Ground level',
            value=well.geometry.z if is_amsl else 0,
            color='green'
        ))
        plot_params.baseLines.append(PlotBaseLine(
            name='Sea level',
            value=0 if is_amsl else - well.geom.z,
            color='blue'
        ))
        ## Stratigraphy

        ## Get the colors of stratigraphies from the QML table
        qml_df = await Qml.get_df(where=Qml.model_name=='GeologicalFormation')
        geological_formation_style = qml_df.iloc[0]
        geological_paint = geological_formation_style.mapbox_paint
        geological_colors_list = geological_paint['fill-color'][2:-1]
        geological_colors_df = pd.DataFrame(
            data=[geological_colors_list[::2],
            geological_colors_list[1::2]]
        ).T
        geological_colors_df.columns = ['id', 'rgb']
        geological_colors_df['color_code'] = geological_colors_df.rgb.apply(
            lambda c: f'#{c[1]:02x}{c[2]:02x}{c[3]:02x}80'
        )
        geological_colors_df['id'] = geological_colors_df.id.astype('int')

        strat_layers_df = await WellStratigraphyLayer.get_df(
            where=WellStratigraphyLayer.well_id==well_id,
            with_related=True
        )
        strat_layers_df = strat_layers_df.merge(
            geological_colors_df[['id', 'color_code']],
            left_on='formation_id',
            right_on='id',
            suffixes=('', '_y')
        )

        if value == 'level_amsl':
            strat_layers_df['top'] -= amsl_shift
            strat_layers_df['bottom'] -= amsl_shift

        if len(strat_layers_df) > 0:
            strat_layers_df.sort_values('order', ascending=True, inplace=True)
            plot_params.bgShapes = [
                PlotBgShape(
                    name=row.avsm_misc_geological_formation_type_formation,
                    valueTop=-row.top,
                    valueBottom=-row.bottom,
                    color=row.color_code
                ) for row in strat_layers_df.itertuples()
            ]

        ## If the depth is unknown, set the barBase as the deepest known level
        try:
            if pd.isna(well.depth):
                query = select(func.min(WellLevel.level)).where(WellLevel.ref_id==well_id)
                async with db_session() as session:
                    min_level = (await session.exec(query)).one()
                if min_level is not None:
                    depth = - min_level + 1  ## Add arbitrary margin
                else:
                    raise NoLevel
            else:
                depth = well.depth
        except NoLevel:
            pass
        else:
            plot_params.barBase = amsl_shift - depth
        return plot_params

    async def get_strat_level(self, level):
        """
        Return the stratigraphy at the given level,
        expressed from the casing top (compatible with level reading)
        """
        if level is None:
            return
        strat_layers_df = await WellStratigraphyLayer.get_df(where=WellStratigraphyLayer.well_id==self.id, with_related=True)
        if len(strat_layers_df) == 0:
            return
        strats = strat_layers_df.loc[(strat_layers_df.top <= abs(level)) & (strat_layers_df.bottom > abs(level))]
        if len(strats) == 0:
            logger.info(f'Level {level} does not match any stratigraphy')
            return
        return strats.iloc[0]

    async def get_info(self) -> list[InfoItem]:
        info = await super(Well, self).get_info()
        if self.depth:
            info.append(InfoItem(key='depth', value=f'{self.depth:.0f} m'))
        else:
            info.append(InfoItem(key='depth', value='?'))
        level_info, trend, level_time, last_level = await self.get_trend()
        if level_info:
            info.append(InfoItem(key='trend', value=trend))
            info.append(InfoItem(key='last level', value=level_info))
            last_level_strat = await self.get_strat_level(last_level)
            if last_level_strat is not None:
                info.append(InfoItem(key='last level stratigraphy',
                    value=last_level_strat.avsm_misc_geological_formation_type_formation))
            info.append(InfoItem(key='last reading', value=level_time))
        if self.type_id:
            info.append(InfoItem(key='well type', value=self.well_type.name))
        if self.status_id:
            info.append(InfoItem(key='status',
                                 value=self.well_status.name
                                       if self.status else 'No status'))
        info.append(InfoItem(key='monitoring',
                             value=self.monitoring_status.name
                                   if self.moni_st_id else 'No monitoring status'))
        info.append(InfoItem(key='in charge',
                             value=self.resident.name
                                   if getattr(self, 'resident', None) else 'No one in charge'))
        if self.casing_top:
            info.append(InfoItem(key='casing top', value=self.casing_top))
        if self.cf:
            info.append(InfoItem(key='correction factor', value=self.cf))
        info.append(InfoItem(key='source',
                             value='CSR Geomatics Studio and others'))
        return info

    async def get_categorized_info(self) -> list[InfoCategory]:
        ## XXX: ideally, related data should acquired automatically with dyn_join_with
        info: list[InfoCategory] = []
        async with db_session() as session:
            query = select(WellStratigraphyLayer).\
                where(WellStratigraphyLayer.well_id==self.id).\
                order_by(WellStratigraphyLayer.order).\
                options(selectinload(WellStratigraphyLayer.formation))
            result = await session.exec(query)
            strat_layers = result.all()
            if len(strat_layers) > 0:
                info.append(InfoCategory(
                    name='Stratigraphy',
                    infoItems=[
                        InfoItem(
                            key=row.formation.formation,
                            value=f'{row.top} m - {row.bottom} m'
                        ) for row in strat_layers
                    ]
                ))
            category: InfoModel
            for category in [WellExtraInfo, WellPumpInfo, WellManagementInfo, WellProductionInfo, WellCrapInfo]:
                query = select(category).\
                    where(category.well_id == self.id).\
                    options(*(joinedload(join) for join in category.selectinload()))
                result = await session.exec(query)
                category_info = result.all()
                if len(category_info) > 0:
                    ## If multiple info for this category, sort by info time
                    ## and assign a different name
                    if len(category_info) > 1:
                        if hasattr(category, 'time'):
                            category_info = sorted(category_info, key='time')
                    for cat_info in category_info:
                        info.append(InfoCategory(
                            name=category.info_category_name,
                            infoItems=cat_info.get_info()
                        ))
        return info

    class Attachments:
        async def images(self) -> list[Attachment]:
            if self.picture:
                return [
                    Attachment(name='picture', path=self.picture)
                ]
            else:
                return []


class WellInfoSource(Model, table=True):
    __tablename__ = 'well_info_source'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water'
        #flask_admin_model_view = 'StratigraphyLayerModelView'

    id: int | None = Field(default=None, primary_key=True)
    name: str
    url: str
    accuracy: float

    def __str__(self):
        return self.name or 'No source name'


class WellPumpType(Model, table=True):
    __tablename__ = 'well_pump_type'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water'
        #flask_admin_model_view = 'StratigraphyLayerModelView'

    id: int | None = Field(default=None, primary_key=True)
    name: str

    def __str__(self):
        return self.name or 'No type name'


class WellPumpPowerSource(Model, table=True):
    __tablename__ = 'well_pump_power_source'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water'
        #flask_admin_model_view = 'StratigraphyLayerModelView'

    id: int | None = Field(default=None, primary_key=True)
    name: str

    def __str__(self):
        return self.name or 'No power source name'


class InfoModel(SQLModel):
    def get_info(self) -> list[InfoItem]:
        infoItems = []
        for key, field in self.model_fields.items():
            if not (getattr(field, 'primary_key', False) or getattr(field, 'foreign_key', False)):
                v = getattr(self, key)
                if v is not None:
                    infoItems.append(InfoItem(
                        key=key,
                        value=str(v)
                    ))
        return infoItems


class WellExtraInfo(Model, InfoModel, table=True):
    __tablename__ = 'well_extra_info'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water'
        #flask_admin_model_view = 'StratigraphyLayerModelView'

    id: int | None = Field(default=None, primary_key=True)
    well_id: int = Field(foreign_key='avsm_water.well.id')
    source: int = Field(foreign_key='avsm_water.well_info_source.id')
    info_source: WellInfoSource = Relationship()
    date: date_
    remarks: str
    well_code: str = Field(sa_type=String(20))
    code_present: bool
    code_written: str = Field(sa_type=String(10))
    description: str
    access: bool
    depth: float
    ## Note: cf factor is manually entered in the Well record, as relative_casing_top - ground_relative_level
    interference: str
    connected_networks: str  ## TODO: relations with Location table

    info_category_name: ClassVar[str] = 'Extra'

    @classmethod
    def selectinload(cls):
        return [
            cls.info_source,
        ]

    # join_with: ClassVar[dict[str, str]] = {
    #     #'well': Well,
    #     'source': WellInfoSource,
    # }


class WellPumpInfo(Model, InfoModel, table=True):
    __tablename__ = 'well_pump_info'
    __table_args__ = water.table_args

    info_category_name: ClassVar[str] = 'Pump'

    class Admin:
        menu = 'Water'
        #flask_admin_model_view = 'StratigraphyLayerModelView'

    id: int | None = Field(default=None, primary_key=True)
    well_id: int = Field(foreign_key='avsm_water.well.id')
    source_id: int = Field(foreign_key='avsm_water.well_info_source.id')
    date: date_
    pump_depth: float
    pump_depth_date: date_
    pump_automatic: bool
    pump_type_id: int = Field(foreign_key='avsm_water.well_pump_type.id')
    pump_brand: str
    pump_model: str
    pump_power: float
    pump_stages: int
    pump_head: float
    pump_installation_date: date_
    pump_power_source_id: int = Field(foreign_key='avsm_water.well_pump_power_source.id')
    pipe_size: float
    delivery_pipe_size: float
    delivery_pipe_material: str = Field(sa_type=String(20))
    discharge_pipe_material: str = Field(sa_type=String(25))
    discharge_pipe_measuring_equipment: str = Field(sa_type=String(25))
    max_pump_capacity_l_per_hour: float
    source: WellInfoSource = Relationship()
    pump_type: WellPumpType = Relationship()
    pump_power_source: WellPumpPowerSource = Relationship()

    @classmethod
    def selectinload(cls):
        return [
            cls.source,
        ]

    # join_with: ClassVar[dict[str, str]] = {
    #     #'well': Well,
    #     'source': WellInfoSource,
    #     'pump_type': WellPumpType,
    #     'pump_power_source': WellPumpPowerSource,
    # }


# class WellHistoricalHarvestExtraInfo(Model, table=True):
#     """
#     From Water Harvest
#     """
#     __tablename__ = 'well_historical_info'
#     __table_args__ = water.table_args

#     class Admin:
#         menu = 'Water'
#         #flask_admin_model_view = 'StratigraphyLayerModelView'

#     id: int | None = Field(default=None, primary_key=True)
#     well_id: int = Field(foreign_key='well.id')
#     old_code: str = Field(sa_type=String(20))
#     old_code_96_97: str = Field(sa_type=String(20))
#     aves_code: str = Field(sa_type=String(20))
#     observation: bool
#     status: str = Field(sa_type=String(20))
#     owner: str = Field(sa_type=String(20))
#     longitude: float
#     latitude: float
#     altitude: float
#     accuracy: float
#     depth: float
#     max_age: str = Field(sa_type=String(20))
#     most_recent_drill_date: date_
#     report_date: date_
#     driller: str = Field(sa_type=String(20))
#     program: str = Field(sa_type=String(20))
#     rig_no: str = Field(sa_type=String(20))
#     diameter: str = Field(sa_type=String(20))
#     drill_type: str = Field(sa_type=String(20))

#     join_with: ClassVar[dict[str, str]] = {
#         #'well': Well,
#         'source': WellInfoSource,
#         'pump_type': WellPumpType,
#         'pump_power_source': WellPumpPowerSource,
#     }


class WellHistoricalHarvestExtraInfo(Model, table=True):
    """
    From Water Harvest
    """
    __tablename__ = 'well_historical_info'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water'
        #flask_admin_model_view = 'StratigraphyLayerModelView'

    id: int | None = Field(default=None, primary_key=True)
    well_id: int = Field(foreign_key='avsm_water.well.id')
    old_code: str = Field(sa_type=String(20))
    old_code_96_97: str = Field(sa_type=String(20))
    aves_code: str = Field(sa_type=String(20))
    observation: bool
    status: str = Field(sa_type=String(20))
    owner: str = Field(sa_type=String(20))
    longitude: float
    latitude: float
    altitude: float
    accuracy: float
    depth: float
    max_age: str = Field(sa_type=String(20))
    most_recent_drill_date: date_
    report_date: date_
    driller: str = Field(sa_type=String(20))
    program: str
    rig_no: str = Field(sa_type=String(20))
    diameter: str = Field(sa_type=String(20))
    drill_type: str = Field(sa_type=String(20))
    well: Well = Relationship()

    @classmethod
    def selectinload(cls):
        return [
            cls.well,
        ]

    # join_with: ClassVar[dict[str, str]] = {
    #     #'well': Well,
    # }


class WellManagementInfo(Model, InfoModel, table=True):
    __tablename__ = 'well_management_info'
    __table_args__ = water.table_args

    info_category_name: ClassVar[str] = 'Management'

    class Admin:
        menu = 'Water'
        #flask_admin_model_view = 'StratigraphyLayerModelView'

    id: int | None = Field(default=None, primary_key=True)
    well_id: int = Field(foreign_key='avsm_water.well.id')
    source_id: int = Field(foreign_key='avsm_water.well_info_source.id')
    date: date_
    casing_depth: float
    has_pump: bool
    pump_functioning: bool
    pack_type: str
    casing_size: float
    casing_material: str = Field(sa_type=String(20))
    slots_depth: str
    casing_bottom: str
    casing_updation: str
    casing_pipe: str
    casing_notes: str
    pump_type_id: int = Field(foreign_key='avsm_water.well_pump_type.id')
    well: Well = Relationship()
    well_info_source: WellInfoSource = Relationship()
    well_pump_type: WellPumpType = Relationship()

    @classmethod
    def selectinload(cls):
        return [
            cls.well,
            cls.well_pump_type,
        ]

    # join_with: ClassVar[dict[str, str]] = {
    #     #'well': Well,
    #     'well_info_source': WellInfoSource,
    #     'well_pump_type': WellPumpType,
    # }


class WellProductionInfo(Model, InfoModel, table=True):
    __tablename__ = 'well_prod_info'
    __table_args__ = water.table_args

    info_category_name: ClassVar[str] = 'Production'

    class Admin:
        menu = 'Water'
        #flask_admin_model_view = 'StratigraphyLayerModelView'

    id: int | None = Field(default=None, primary_key=True)
    well_id: int = Field(foreign_key='avsm_water.well.id')
    source_id: int = Field(foreign_key='avsm_water.well_info_source.id')
    date: date_
    ## TODO: start from Wcap_H column
    capacity_info: str
    capacity: float  ## in l/h
    output: float
    water_usage: str  ## Free text, calling for proper structure
    interconnections: str  ## Free text, calling for proper structure
    average_daily_extraction: str
    max_daily_output: float
    tank_storage_capacity: str
    daily_yield_capacity: float  ## In l/day?
    discharge: str
    average_pumping_daily: float  ## hours
    well: Well = Relationship()
    well_info_source: WellInfoSource = Relationship()

    @classmethod
    def selectinload(cls):
        return [
            cls.well,
            cls.well_info_source,
        ]

    # join_with: ClassVar[dict[str, str]] = {
    #     #'well': Well,
    #     'well_info_source': WellInfoSource,
    # }


class WellContactInfo(Model, table=True):
    __tablename__ = 'well_contact_info'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water'

    id: int | None = Field(default=None, primary_key=True)
    well_id: int = Field(foreign_key='avsm_water.well.id')
    caretaker: str = Field(sa_type=String(25))
    contact_person: str = Field(sa_type=String(25))
    contact_phone: str = Field(sa_type=String(30))
    contact_mobile: str = Field(sa_type=String(30))
    contact_mail: str = Field(sa_type=String(30))
    well: Well = Relationship()

    @classmethod
    def selectinload(cls):
        return [
            cls.well,
        ]

    # join_with: ClassVar[dict[str, str]] = {
    #     #'well': Well,
    # }


class WellCrapInfo(Model, InfoModel, table=True):
    __tablename__ = 'well_crap_info'
    __table_args__ = water.table_args

    info_category_name: ClassVar[str] = 'Other'

    class Admin:
        menu = 'Water'

    id: int | None = Field(default=None, primary_key=True)
    well_id: int = Field(foreign_key='avsm_water.well.id')
    source_id: int = Field(foreign_key='avsm_water.well_info_source.id')
    source_note: str
    date: date_
    ## TODO: use free key/value field
    relative_ground_level: float
    relative_casing_top_level: float
    pump_maintenance: str
    additional_pump_depth_records: str
    pump_water_level: str
    water_level_extra_info: str
    pump_notes: str
    pump_transformer_code: str
    pump_transformer_name: str
    pump_transformer_longitude: float
    pump_transformer_latitude: float
    pump_transformer_elevation: float
    pump_transformer_connections: str
    pump_transformer_additional_info: str
    ## An extra field for storing all kind of random crap
    misc: dict = Field(sa_type=MutableDict.as_mutable(HSTORE))
    well_info_source: WellInfoSource = Relationship()

    @classmethod
    def selectinload(cls):
        return [cls.well_info_source]

    # join_with: ClassVar[dict[str, str]] = {
    #     #'well': Well,
    #     'well_info_source': WellInfoSource,
    # }


class WellStratigraphyLayer(Model, table=True):
    __tablename__ = 'well_strat_layer'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'StratigraphyLayerModelView'

    id: int | None = Field(default=None, primary_key=True)
    well_id: int = Field(foreign_key='well.id')
    ## The order of this layer for that well, from top to bottom
    order: int
    formation_id: int = Field(foreign_key='avsm_misc.geological_formation_type.id')
    top: float
    bottom: float
    formation: GeologicalFormationType = Relationship()

    @classmethod
    def selectinload(cls):
        return [cls.formation]

    # join_with: ClassVar[dict[str, str]] = {
    #     #'well': Well,
    #     'formation': GeologicalFormationType,
    # }


class WellRemoteMonitoring(Model, table=True):
    __tablename__ = 'well_remote_monitoring'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'WellRemoteMonitoringModelView'

    id: int | None = Field(default=None, primary_key=True)
    name: str
    ttn_key: str
    mac_id: str
    well_id: int = Field(foreign_key='avsm_water.well.id')
    enabled: bool

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<models.WellRemoteMonitoring {self.name:s}>'.format(self=self)


class WellLevel(TimePlottableModel, table=True):
    __tablename__ = 'well_level'
    __table_args__ = water.table_args

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'WellLevelsModelView'

    id: int | None = Field(default=None, primary_key=True)
    level: float
    dry: bool = False
    ref_id: int = Field(foreign_key='avsm_water.well.id')
    ref: 'Well' = Relationship()

    values: ClassVar[list[dict[str: str | int]]] = [
        {
            'name': 'level',
            'title': 'Level (ref. casing top)',
            'agg': 'mean',
            'unit': 'm',
            'round': 2,
            'chartType': 'bar',
            'chartColor': 'blue',
        },
        {
            'name': 'level_amsl',
            'title': 'Level (ref. mean sea level - AMSL)',
            'agg': 'mean',
            'unit': 'm',
            'round': 2,
            'chartType': 'bar',
            'chartColor': 'blue',
        },
    ]

    def as_absolute(self, well) -> float:
        """
        Absolute level, from sea level
        """
        if not well.casing_top:
            raise NoCasingTop
        return self.level + well.casing_top

    @classmethod
    async def get_as_dataframe(cls, item, with_only_columns=None, *args, **kwargs):
        if with_only_columns is None:
            with_only_columns = ['level', 'dry']
        df = await super().get_as_dataframe(item, with_only_columns=with_only_columns,
                                            *args, **kwargs)
        df['level_amsl'] = item.elevation + (item.cf or 0) + df['level']
        return df


class WellLevelBakset(Basket):
    name = 'Well levels'


class AddLevelPlugin(ActionPlugin):
    async def execute(self,
                      user: UserRead,
                      features: dict[str, list],
                      params: list,
                      form_fields: list[FormFieldInput]) -> ActionResult:
        vals = {
            f.name: f.value
            for f in form_fields
        }
        date = dateutil.parser.parse(vals['Date']).astimezone(get_localzone()).date()
        time = dateutil.parser.parse(vals['Time']).time()
        level = -abs(float(vals['Level']))
        async with db_session() as session:
            session.add(WellLevel(
                ref_id=features['avsm_water.well'][0],
                time=datetime.combine(date, time).replace(tzinfo=None),
                level=level,
                dry=vals['Dry'] == 'true',
            ))
            await session.commit()
        return ActionResult(message='Well level saved')

def strat_layer_for_user(row):
    """
    Utility function to create a user readable description of the stratigraphy layers
    """
    return f'{row.avsm_misc_geological_formation_type_formation} ({row.top} m - {row.bottom} m)'


#    async def get_date(self):
#        return datetime.now().date().isoformat()
#
#    async def get_time(self):
#        return datetime.now().time().isoformat('seconds')
#
#    async def get_last_level_rounded(self):
#        last_level = await WellLevel.query.where(WellLevel.ref_id==self.id).\
#            order_by(WellLevel.time.desc()).limit(1).gino.first()
#        if last_level:
#            return '{:.3f}'.format(last_level.level)
#
#    async def is_dry(self):
#        return None
#
#    fields = {ActionPlugin
#        'Date': ('date', get_date),
#        'Time': ('time', get_time),
#        'Level': ('number', get_last_level_rounded),
#        'Dry': ('checkbox', is_dry),
#    }
#
#    async def process(self, form_fields):
#        ## Convert the date (set at midnight utc) to timezone
#        date = dateutil.parser.parse(form_fields['Date']).astimezone(get_localzone()).date()
#        time = dateutil.parser.parse(form_fields['Time']).time()
#        level = -abs(float(form_fields['Level']))
#        await WellLevel.create(
#            ref_id=self.id,
#            time=datetime.combine(date, time).replace(tzinfo=None),
#            level=level,
#            dry=form_fields.get('Dry', False),
#        )
#        return f'Level for well {self.caption} saved'


add_level = AddLevelPlugin(
    name='Add a new level reading',
    stores=['avsm_water.well'],
    roles=['well_level_operator'],
    icon='addchart',
    form_fields=[
        FormField(
            name='Date',
            type='date'
        ),
        FormField(
            name='Time',
            type='time'
        ),
        FormField(
            name='Level',
            type='number'
        ),
        FormField(
            name='Dry',
            type='checkbox'
        ),
    ]
)


download_all_as_csv = DownloadCSVPlugin(
    name='Raw well levels as CSV',
    stores=['avsm_water.well'],
    icon='storage',
)
