import logging
from datetime import date
from typing import ClassVar, Any

from sqlmodel import Field, Relationship

from gisaf.models.models_base import Model
from gisaf.models.geo_models_base import TimePlottableModel
from gisaf.baskets import Basket
from gisaf.plugins import DownloadCSVPlugin

from avgs_gisaf_extra_models.models_base import (GeoPointModelWithResident,
                                                 Location, Resident)
from gisaf.models.info_item import InfoItem

from ..metadata import misc

logger = logging.getLogger(__name__)

# weather_station_id_seq = db.Sequence('weather_station_id_seq')
# weather_station_level_id_seq = db.Sequence('weather_station_level_id_seq')


class WeatherStationProvider(Model, table=True):
    __tablename__ = 'weather_station_provider'
    __table_args__ = misc.table_args

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'MyModelView'

    id: int | None = Field(default=None, primary_key=True)
    name: str
    url: str
    api_key: str
    weather_stations: list['WeatherStation'] = Relationship(back_populates='provider')

    def __repr__(self):
        return f'<models.WeatherStationProvider {self.name:s}>'

    def __str__(self):
        return self.name


class WeatherStation(GeoPointModelWithResident, table=True):
    __tablename__ = 'weather_station'
    __table_args__ = misc.table_args

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'WithLevelsAndLocationModelView'

    id: int | None = Field(default=None, primary_key=True)
    since: date  ## Year, should change to date (default: 1/1/year, YYYY-MM-DD)
    provider_id: int = Field(foreign_key='avsm_misc.weather_station_provider.id')
    provider: 'WeatherStationProvider' = Relationship(back_populates='weather_stations')
    provider_unit_id: str
    levels: list['WeatherStationLevels'] = Relationship(back_populates='ref')
    offsets: list['WeatherStationLevelsRainOffset'] = Relationship(back_populates='ref')

    location: Location = Relationship()
    resident: Resident = Relationship()

    icon: ClassVar[str] = 'avsm_misc-weather_station'
    #symbol = '\ue007'

    @classmethod
    def selectinload(cls):
        return [cls.provider, cls.location, cls.resident]


    # join_with: ClassVar[dict[str, Any]] = {
    #     'provider': WeatherStationProvider,
    # }

    async def get_info(self) -> list[InfoItem]:
        return [
            InfoItem(key='source', value='CSR Geomatics Studio')
        ]


class WeatherStationLevelsRainOffset(Model, table=True):
    """
    For Weather Undergound WS (in CSR, Kalpana): the total_rainfall counter
    is blocked at 6000 mm.
    Refs: https://redmine.auroville.org.in/issues/13259
    create table avsm_misc.weather_station_levels_rain_offset (
        id serial PRIMARY KEY, ref_id INT, since DATE, "offset" FLOAT, is_reset BOOLEAN);
    alter table avsm_misc.weather_station_levels_rain_offset
        add constraint  weather_station_levels_rain_offset_ref_id_pk
            foreign key(ref_id) references avsm_misc.weather_station(id)
            on update cascade on delete cascade;
    ALTER TABLE avsm_misc.weather_station_levels_rain_offset OWNER TO avgis;
    """

    __tablename__ = 'weather_station_levels_rain_offset'
    __table_args__ = misc.table_args

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'MyModelView'

    id: int | None = Field(default=None, primary_key=True)
    ref_id: int = Field(foreign_key='avsm_misc.weather_station.id')
    ref: WeatherStation = Relationship(back_populates='offsets')
    since: date  ## Year, should change to date (default: 1/1/year, YYYY-MM-DD)
    offset: float
    is_reset: bool


class WeatherStationLevels(TimePlottableModel, table=True):
    __tablename__ = 'weather_station_levels'
    __table_args__ = misc.table_args

    class Admin:
        menu = 'Water'
        flask_admin_model_view = 'WeatherStationLevelsModelView'

    id: int | None = Field(default=None, primary_key=True)
    temperature: float
    pressure: float
    sun_duration: float
    humidity: float
    wind_speed: float
    wind_direction: float
    rainfall: float
    daily_rainfall: float
    total_rainfall: float
    solarradiation: float
    uv: float
    ref_id: int = Field(foreign_key='avsm_misc.weather_station.id')
    ref: WeatherStation = Relationship(back_populates='levels')

    resampling_args: ClassVar[dict[str, dict[str, dict[str, str]]]] = {
        ## Resampling rain fall with the same base as manual rain gauges in India
        ## See https://redmine.auroville.org.in/issues/9220
        'rainfall': {
            'D': {
                #'base': 8.5 / 24,
                'offset': '8.5h',
                'comment': "Base of resampling set at 8:30AM, "
                           "Indian's convention for manual raingauges",
                'loffset': '1D'
            },
        },
    }

    values: ClassVar[list[dict[str, str]]] = [
        {
            'name': 'rainfall',
            'title': 'Rainfall',
            'unit': 'mm',
            'agg': 'sum',
            'chartType': 'bar',
            'chartColor': 'blue',
        },
        {
            'name': 'temperature',
            'title': 'Temperature',
            'unit': '°C',
            'agg': 'mean',
            'chartColor': '#408080',
        },
        {
            'name': 'pressure',
            'title': 'Pressure',
            'unit': 'mb',
            'agg': 'mean',
            'chartColor': '#405050',
        },
        {
            'name': 'humidity',
            'title': 'Humidity',
            'unit': '%',
            'agg': 'mean',
            'chartColor': '#4040F0',
        },
        {
            'name': 'wind_speed',
            'title': 'Wind speed',
            'unit': 'km/h',
            'agg': 'mean',
            'chartColor': '#4040F0',
        },
        {
            'name': 'wind_direction',
            'title': 'Wind direction',
            'unit': '°',
            'agg': 'mean',
            'chartColor': '#4040F0',
        },
        {
            'name': 'solarradiation',
            'title': 'Solar radiation',
            'unit': 'W/sq. m',
            'agg': 'mean',
            'chartColor': '#F04060',
        },
        {
            'name': 'uv',
            'title': 'UV',
            'unit': '',
            'agg': 'mean',
            'chartColor': '#F00000',
        }
    ]

    @classmethod
    async def get_as_dataframe(cls, item, with_only_columns=None, **kwargs):
        if with_only_columns is None:
            with_only_columns = [val['name'] for val in cls.values]
        convert_rainfall = 'rainfall' in with_only_columns

        if convert_rainfall:
            columns = with_only_columns.copy()
            columns.remove('rainfall')
            columns.append('total_rainfall')
        else:
            columns = with_only_columns

        df = await super().get_as_dataframe(item, with_only_columns=columns, **kwargs)

        if convert_rainfall:
            ## XXX: maybe this should be: df['rainfall'] = df.total_rainfall.fillna(method='pad').diff()
            df['rainfall'] = df['total_rainfall'].diff()

            ## Fix for hickups: single noise values, easily seen as the cumulative
            ## rainfall decreases. This might happen for a wide variety of reasons.
            df1 = df.drop(df[df["rainfall"]<0].index)
            ## TODO: this works only for offending values are on the down side of the curve.
            ## For upper side noise, the noisy value is the sample just before the detected negative diff
            df = df1
            df.drop(columns=['total_rainfall'], inplace=True)

        if 'wind_speed' in df.columns:
            ## Convert wind speed from m/s to km/h
            df['wind_speed'] = df['wind_speed'] * 3.6

        return df


class WeatherStationLevelsBasket(Basket):
    name: ClassVar[str] = 'Weather station levels'


download_all_as_csv = DownloadCSVPlugin(
    name='Raw weather station levels as CSV',
    stores=['avsm_misc.weather_station'],
    icon='storage',
)
