from typing import ClassVar

from sqlalchemy import String
from sqlmodel import Field

from gisaf.models.geo_models_base import GeoPointModel, GeoLineModel
from gisaf.models.info_item import InfoItem

from ..metadata import misc, avsm

class OpticalFiber(GeoLineModel, table=True):
    __tablename__ = 'optical_fiber'
    __table_args__ = misc.table_args

    icon: ClassVar[str] = 'avsm_misc-optical_fiber'
    attribution: ClassVar[str] = 'Aurinoco'

    class Admin:
        menu = 'Infrastructure'

    id: int | None = Field(default=None, primary_key=True)
    cores: int
    start_code: str = Field(sa_type=String(6)) # type: ignore
    end_code: str = Field(sa_type=String(6)) # type: ignore
    cable_code: str = Field(sa_type=String(15)) # type: ignore
    batch_nbr: str = Field(sa_type=String(5)) # type: ignore
    start_n: int
    end_n: int
    manufactu: str = Field(sa_type=String(15)) # type: ignore

    synonyms: ClassVar[dict[str, str]] = {
        'batch_#': 'batch_nbr',
        'start_': 'start_n',
        'end_': 'end_n',
    }

    def __str__(self):
        return f'{self.cores} cores, {self.start_code} to {self.end_code}'

    def __repr__(self):
        return f'<models.OpticalFiber {self.id:s}>'


    @classmethod
    async def get_popup(cls, df):
        return 'Optical fiber: ' + df['cable_code'].fillna('?') + ', ' + \
            df.cores.astype('Int64').astype('U').replace('nan', '?') + ' cores'

    @classmethod
    async def get_properties(cls, df):
        return {
            'cores': df['cores'].fillna(0).astype('int64').astype('U')
        }

    @property
    def caption(self):
        if self.cores:
            return f"Optical fiber {self.cable_code or 'unknown'} ({self.cores} cores)"
        else:
            return f"Optical fiber {self.cable_code or 'unknown'}"

    async def get_feature_properties(self):
        ## DEPRECATED
        return {
            'cores': self.cores
        }

    async def get_info(self) -> list[InfoItem]:
        info = await super(OpticalFiber, self).get_info()
        info.append(InfoItem(key='cores', value=self.cores))
        info.append(InfoItem(key='start code', value=self.start_code))
        info.append(InfoItem(key='end code', value=self.end_code))
        info.append(InfoItem(key='cable code', value=self.cable_code))
        info.append(InfoItem(key='batch #', value=self.batch_nbr))
        info.append(InfoItem(key='start n', value=self.start_n))
        info.append(InfoItem(key='end n', value=self.end_n))
        info.append(InfoItem(key='manufacturer', value=self.manufactu))
        return info

class OpticalFiberPoint(GeoPointModel, table=True):
    __tablename__ = 'optical_fiber_point'
    __table_args__ = misc.table_args

    icon: ClassVar[str] = 'avsm_misc-optical_fiber'
    attribution: ClassVar[str] = 'Aurinoco'

    class Admin:
        menu = 'Infrastructure'

    id: int | None = Field(default=None, primary_key=True)
    code: str = Field(sa_type=String(10)) # type: ignore
    type: str = Field(sa_type=String(6)) # type: ignore

    synonyms: ClassVar[dict[str, str]] = {
        'OFC_Code': 'code',
        'Type': 'type',
    }

    def __str__(self):
        return f'{self.code}'

    def __repr__(self):
        return f'<models.OpticalFiberPoint {self.id:s}>'

    async def get_info(self) -> list[InfoItem]:
        info = await super(OpticalFiberPoint, self).get_info()
        info.append(InfoItem(key='code', value=self.code))
        info.append(InfoItem(key='type', value=self.type))
        return info