from typing import ClassVar
from sqlmodel import Field, Relationship

from gisaf.models.models_base import Model
from gisaf.models.geo_models_base import GeoLineModel, GeoPointModel
from gisaf.models.info_item import InfoItem

from ..metadata import misc

# telephone_box_type_id_seq = db.Sequence('telephone_box_type_id_seq')
# telephone_box_mtype_id_seq = db.Sequence('telephone_box_mtype_id_seq')
# telephone_box_id_seq = db.Sequence('telephone_box_id_seq')
# telephone_cable_type_id_seq = db.Sequence('telephone_cable_type_id_seq')
# telephone_line_id_seq = db.Sequence('telephone_line_id_seq')


class TelephoneBoxType(Model, table=True):
    __tablename__ = 'telephone_box_type'
    __table_args__ = misc.table_args

    class Admin:
        menu = 'Infrastructure'

    id: int | None = Field(default=None, primary_key=True)
    name: str
    boxes: list['TelephoneBox'] = Relationship(back_populates='type')

    def __str__(self):
        return self.name or ''

    def __repr__(self):
        return '<models.TelephoneBoxType {self.name:s}>'.format(self=self)


class TelephoneBoxMountingType(Model, table=True):
    __tablename__ = 'telephone_box_mtype'
    __table_args__ = misc.table_args

    class Admin:
        menu = 'Infrastructure'

    id: int | None = Field(default=None, primary_key=True)
    name: str
    boxes: list['TelephoneBox'] = Relationship(back_populates='mounting_type')

    def __str__(self):
        return self.name or ''

    def __repr__(self):
        return '<models.TelephoneBoxMountingType {self.name:s}>'.format(self=self)


class TelephoneBox(GeoPointModel, table=True):
    __tablename__ = 'telephone_box'
    __table_args__ = misc.table_args

    class Admin:
        menu = 'Infrastructure'

    z_index: ClassVar[int] = 430
    icon: ClassVar[str] = 'avsm_misc-telephone_box'
    attribution: ClassVar[str] = 'AVTS'

    id: int | None = Field(default=None, primary_key=True)
    info: str
    box_number: str
    type_id: int = Field(foreign_key='avsm_misc.telephone_box_type.id')
    type: 'TelephoneBoxType' = Relationship(back_populates='boxes')
    mtype_id: int = Field(foreign_key='avsm_misc.telephone_box_mtype.id')
    mounting_type: 'TelephoneBoxMountingType' = Relationship(back_populates='boxes')

    @classmethod
    def selectinload(cls):
        return [
            cls.type,
            cls.mounting_type,
        ]

    # join_with: ClassVar[dict[str, str]] = {
    #     'type': TelephoneBoxType,
    #     'mounting_type': TelephoneBoxMountingType,
    # }

    def get_main_box(self):
        """
        Get the "root" of the box.
        Eg. the box "1" for the box's c_number "1.2.3.4"
        :return:
        """
        return

    async def get_info(self) -> list[InfoItem]:
        info = await super(TelephoneBox, self).get_info()
        info.append(InfoItem(key='box number', value=self.box_number))
        if self.type_id:
            info.append(InfoItem(key='type', value=str(self.type)))
        if self.mtype_id:
            info.append(InfoItem(key='mounting type', value=str(self.mounting_type)))
        info.append(InfoItem(key='source', value="Johan - many thanks to him!"))
        ## XXX: hide info for time being as we're not sure about confidentiality
        #info['info'] = self.info
        return info

    @classmethod
    async def get_popup(cls, df):
        return 'Telephone box: ' + df['box_number']

    @classmethod
    async def get_properties(cls, df):
        return {
            'class': df['mtype_id'].fillna(0).to_numpy('int64').astype('U')
        }

    @property
    def caption(self):
        return f'Telephone box {self.box_number}'
        ## XXX: hide info for time being as we're not sure about confidentiality
        #if self.info:
        #    return "{:s} ({:s})".format(self.box_number, self.info)
        #else:
        #    return self.box_number

    async def get_feature_properties(self):
        ## DEPRECATED
        if self.mtype_id:
            return {
                'class': str(self.mtype_id)
            }
        else:
            return {}


class TelephoneCableType(Model, table=True):
    __tablename__ = 'telephone_cable_type'
    __table_args__ = misc.table_args

    class Admin:
        menu = 'Infrastructure'

    id: int | None = Field(default=None, primary_key=True)
    name: str

    def __str__(self):
        return self.name or ''

    def __repr__(self):
        return '<models.TelephoneCableType {self.name:s}>'.format(self=self)


class TelephoneLine(GeoLineModel, table=True):
    __tablename__ = 'telephone_line'
    __table_args__ = misc.table_args

    class Admin:
        menu = 'Infrastructure'

    z_index: ClassVar[int] = 429
    attribution: ClassVar[str] = 'AVTS'
    icon: ClassVar[str] = 'avsm_misc-telephone_line'

    id: int | None = Field(default=None, primary_key=True)
    info: str
    c_number: str
    no_of_pair: int
    type_id: int = Field(foreign_key='avsm_misc.telephone_cable_type.id')
    type: TelephoneCableType = Relationship()
    m_feed_id: int = Field(foreign_key='avsm_misc.telephone_box.id')
    main_feed: TelephoneBox = Relationship()

    @classmethod
    def selectinload(cls):
        return [
            cls.type,
            cls.main_feed,
        ]

    # join_with: ClassVar[dict[str, str]] = {
    #     'type': TelephoneCableType,
    #     'main_feed': TelephoneBox,
    # }

    @classmethod
    async def get_popup(cls, df):
        return 'Telephone line: ' + df['c_number'] + ' (' + df['info'] + ')'

    @classmethod
    async def get_properties(cls, df):
        return {
            'class': df['m_feed_id'].to_numpy('U')
        }

    @property
    def caption(self):
        if self.info:
            return f"Telephone line {self.c_number} ({self.info})"
        else:
            return f"Telephone line {self.c_number}"

    def get_main_feed(self):
        """
        Get the "root" of the line.
        Eg. the line "1" for the line's c_number "1.2.3.4"
        :return:
        """
        return

    async def get_info(self) -> list[InfoItem]:
        info = await super(TelephoneLine, self).get_info()
        info.append(InfoItem(key='info', value=self.info))
        if self.type_id:
            info.append(InfoItem(key='type', value=str(self.type)))
        info.append(InfoItem(key='c_number', value=self.c_number))
        info.append(InfoItem(key='number of pairs', value=self.no_of_pair))
        info.append(InfoItem(key='main feed', value=str(self.main_feed)))
        info.append(InfoItem(key='source', value="Johan - many thanks to him!"))
        return info

    async def get_feature_properties(self):
        ## DEPRECATED
        return {
            'class': str(self.m_feed_id)
        }
