from typing import Any, Literal
from sqlmodel import Field, Relationship
from sqlalchemy.orm import QueryableAttribute

from gisaf.models.models_base import Model
from gisaf.models.geo_models_base import GeoPointModel, GeoPointZModel

from .metadata import avsm

# info_source_id_seq = db.Sequence('info_source_id_seq')
# resident_id_seq = db.Sequence('resident_id_seq')
# location_id_seq = db.Sequence('location_id_seq')


class InfoSource(Model, table=True):
    __tablename__ = 'info_source'
    __table_args__ = avsm.table_args

    class Admin:
        menu = 'Other'
        flask_admin_model_view = 'MyModelViewWithName'

    id: int | None = Field(default=None, primary_key=True)
    name: str
    address: str

    def __str__(self):
        return '{self.name:s} @ {self.address:s}'.format(self=self)

    def __repr__(self):
        return '<models.InfoSource {self.name:s} @ {self.address:s}>'.format(self=self)


class Resident(Model, table=True):
    """
    Fields from CSV: id,loc_id,res,ll_phone,mob_phone,email
    """
    __tablename__ = 'resident'
    __table_args__ = avsm.table_args

    class Admin:
        menu = 'Other'
        flask_admin_model_view = 'MyModelViewWithName'

    id: int | None = Field(default=None, primary_key=True)
    loc_id: int = Field(foreign_key='avsm.location.id')
    location: 'Location' = Relationship(back_populates='residents')

    name: str  ## XXX: from Asyncto?
    ll_phone: str
    mob_phone: str
    email: str

    @classmethod
    def selectinload(cls):
        return [
            cls.location # type: ignore
        ]

    def __str__(self):
        return '{self.name:s} @ {self.location.name}'.format(self=self) # type: ignore

    def __repr__(self):
        return '<models.Resident {self.name:s} @ {self.location.name}>'.format(self=self) # type: ignore


class Location(Model, table=True):
    __tablename__ = 'location'
    __table_args__ = avsm.table_args

    class Admin:
        menu = 'Other'
        flask_admin_model_view = 'MyModelViewWithName'

    id: int | None = Field(default=None, primary_key=True)
    name: str
    residents: list[Resident] = Relationship(back_populates='location')

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<models.Location {self.name:s}>'.format(self=self)


class GeoPointWithLocationModel(GeoPointModel):
    loc_id: int = Field(foreign_key='avsm.location.id')

    @classmethod
    def selectinload(cls) -> list[Literal['*'] | QueryableAttribute[Any]]:
        return [
            cls.location, # type: ignore
        ]

    @classmethod
    async def get_popup(cls, df):
        return cls.__name__ + ': ' + df['avsm_location_name'] + ' (' + df.index.astype('U') + ')'

    @property
    def caption(self) -> str:
        """
        Return user friendly name (used in menu, etc)
        :return:
        """
        return f'{self.location} ({self.id:d})'.format(self=self) # type: ignore


class GeoPointModelWithResident(GeoPointWithLocationModel):
    """
    Base class for models with assigned to a resident.
    Subclasses should include, so that df.avsm_resident_name is populated:
    resident: Resident = Relationship()
    """
    res_id: int = Field(foreign_key='avsm.resident.id')

    @classmethod
    def selectinload(cls) -> list[Literal['*'] | QueryableAttribute[Any]]:
        return [
            cls.location, # type: ignore
            cls.resident, # type: ignore
        ]

    @classmethod
    async def get_popup(cls, df):
        return cls.__name__ + ': ' + df.avsm_resident_name.fillna('<No resident>') + ' @ ' + df['avsm_location_name'] + ' (' + df.index.astype('U') + ')' # type: ignore

    @property
    def caption(self) -> str:
        """
        Return user friendly name (used in menu, etc)
        :return:
        """
        if hasattr(self, 'resident') and self.resident is not None:
            return f'{self.resident.name} @ {self.location} ({self.id})' # type: ignore
        else:
            return f'<No resident> @ {self.location} ({self.id})' # type: ignore


class GeoPointZModelWithResident(GeoPointZModel, GeoPointModelWithResident):
    # __abstract__ = True
    pass
