"""
Link with FAMC asset management APIs
"""
import asyncio
import logging
from functools import partial

from sqlalchemy import and_

from redminelib import Redmine, exceptions

from gisaf.models.tags import Tags as TagsModel, TagKey as TagKeyModel
from gisaf.registry import registry
from gisaf.plugins import (
    TagPlugin, ActionPlugin,
    ActionResult, ActionResults,
    TaggedLayer, TaggedFeature,
    create_tags
)
from gisaf.config import conf

logger = logging.getLogger('Redmine plugin')


ticket_url = TagPlugin(
    domain='redmine',
    key='ticket',
    stores_by_re=[''],  ## Match everything
    link='https://redmine.auroville.org.in/issues/{value}',
)


class CreateTicket(ActionPlugin):
    redmine = None
    project = None

    async def execute(self, user, features, params, formFields) -> ActionResults:
        ## Connect to the Redmine server with the first call of this function.
        if not self.redmine:
            ## XXX: case of lost connection?
            ## Some validation of the config
            if 'redmine' not in conf.plugins:
                logger.warn('No plugin configuration')
                return ActionResults(message='Cannot contact the Redmine server (nothing configured)')
            redmine_user = conf.plugins['redmine'].get('user')
            password = conf.plugins['redmine'].get('password')
            baseUrl = conf.plugins['redmine'].get('baseUrl')
            if not (baseUrl and redmine_user and baseUrl):
                logger.warn('User and password must be defined')
                return ActionResults(message='Cannot contact the Redmine server (no user and password provided)')
            self.redmine = Redmine(baseUrl, username=redmine_user, password=password)

        if not self.project:
            loop = asyncio.get_running_loop()
            try:
                self.project = await loop.run_in_executor(None, self.redmine.project.get, 'map-features')
            except exceptions.BaseRedmineError as err:
                return ActionResults(message=f'Cannot contact the Redmine server ({err.args[0]})')
            except:
                return ActionResults(message=f'Cannot contact the Redmine server (Server or network down?)')

        issues = []
        results = ActionResults(actionResults=[])
        result_layers = []
        for store, ids in features.items():
            result_layer = TaggedLayer(store=store, taggedFeatures=[])
            model = registry.stores.loc[store, 'model']
            store_features = await model.get_df(where=model.id.in_(ids), geom_as_ewkt=True)
            existing_tags = await TagsModel.get_df(where=and_(TagsModel.ref_id.in_(ids), TagsModel.store==store), geom_as_ewkt=True)
            for id in ids:
                ## Check if such a tag already exists
                if existing_tags.loc[existing_tags.id==id].tags.apply(lambda t: ticket_url.full_key in t).any():
                    ## There's already a tag for this: pass
                    continue
                ## Create a new issue
                loop = asyncio.get_running_loop()
                try:
                    format = {
                        'user': user.username,
                        'store': store,
                        'id': id
                    }
                    subject = conf.plugins['redmine']['subject'].strip().format(**format)
                    description = conf.plugins['redmine']['description'].strip().format(**format)

                    ## Redmine python is sync, using loop.run_in_executor
                    issue = await loop.run_in_executor(
                        None,
                        partial(
                            self.redmine.issue.create,
                            project_id=self.project.id,
                            subject=subject,
                            description=description,
                        )
                    )
                except Exception as err:
                    raise err
                issues.append(issue)

        result = await create_tags(
            features,
            keys=[ticket_url.full_key],
            values=[
                str(issue.id) for issue in issues
            ]
        )

        if len(issues) == 0:
            results.message = f'No new Redmine issue created (please check existing tags)'
        else:
            results.message = 'Created issue{} {}'.format(
                '' if len(issues) == 1 else 's',
                ', '.join([str(issue.id) for issue in issues])
            )
        return results


create_ticket = CreateTicket(
    name='Create Redmine ticket',
    icon='add_alert',
    stores_by_re=[''],
    roles=[conf.plugins['redmine'].get('role_create')]
)